@@ -7,22 +7,22 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from datetime import datetime
 from types import TracebackType
-from typing import Optional
+from typing import Any, Dict, List, Optional, Tuple
 
 import libsbml as libsbml
 
 from .utils import convert_id_to_sbml, warning_on_one_line
 
-warnings.formatwarning = warning_on_one_line
+warnings.formatwarning = warning_on_one_line  # type: ignore
 
 
 @dataclass
 class Model:
     """Meta-info container for model."""
 
-    sbo: str = None
-    id: str = None
-    name: str = None
+    sbo: Optional[str] = None
+    id: Optional[str] = None
+    name: Optional[str] = None
     units: dict = field(default_factory=dict)
     compartments: dict = field(default_factory=dict)
     notes: dict = field(default_factory=dict)
@@ -31,8 +31,8 @@ class Model:
 class BaseModel(ABC):
     """Abstract model class."""
 
-    def __init__(self, meta_info: dict = None):
-        default_meta_info = {
+    def __init__(self, meta_info: Dict[str, Any] = None) -> None:
+        default_meta_info: Dict[str, Any] = {
             "sbo": "SBO:0000004",  # modelling framework
             "id": f"modelbase-model-{datetime.now().date().strftime('%Y-%m-%d')}",
             "name": "modelbase-model",
@@ -56,10 +56,10 @@ class BaseModel(ABC):
         }
         if meta_info is not None:
             default_meta_info.update(meta_info)
-        self.meta_info = {"model": Model(**default_meta_info)}
-        self._ids = set()
+        self.meta_info: Dict[str, Any] = {"model": Model(**default_meta_info)}
+        self._ids: set[str] = set()
 
-    def __enter__(self):
+    def __enter__(self) -> Any:
         """Enter the context manager.
 
         Returns
@@ -74,14 +74,14 @@ class BaseModel(ABC):
         exception_type: Optional[BaseException],
         exception_value: Optional[BaseException],
         exception_traceback: Optional[TracebackType],
-    ):
+    ) -> None:
         """Exit the context manager.
 
         Restores any changes made to the model structure.
         """
         self.__dict__ = self._copy.__dict__
 
-    def copy(self) -> BaseModel:
+    def copy(self) -> Any:
         """Create a deepcopy of the model.
 
         Returns
@@ -91,31 +91,44 @@ class BaseModel(ABC):
         """
         return copy.deepcopy(self)
 
-    def _check_for_existence(self, *, name: str, check_against: list[str], candidates: list[str]):
-        check_against = set(check_against)
-        check_against.add("time")
-        for i in set(candidates).difference(check_against):
-            warnings.warn(f"Could not find {i} for {name}")
+    def _check_for_existence(
+        self,
+        *,
+        name: str,
+        check_against: List[str],
+        candidates: List[str],
+        element_type: str,
+        raise_on_missing: bool = False,
+    ) -> None:
+        to_check = set(check_against)
+        to_check.add("time")
+        for i in set(candidates).difference(to_check):
+            if raise_on_missing:
+                raise KeyError(f"Could not find {element_type} {i} for {name}")
+            else:
+                warnings.warn(f"Could not find {element_type} {i} for {name}")
 
-    def _check_and_insert_ids(self, ids: list[str]):
+    def _check_and_insert_ids(self, ids: List[str], context: str) -> None:
         for id_ in ids:
             if id_ in self._ids:
-                warnings.warn(f"Id {id_} already exists in the model. This might lead to variable shading!")
+                warnings.warn(
+                    f"{context}: Id {id_} already exists in the model. This might lead to variable shading!"
+                )
             self._ids.add(id_)
 
-    def _update_ids(self, ids: dict[str, str]):
+    def _update_ids(self, ids: Dict[str, str]) -> None:
         for k, v in ids.items():
             self._ids.remove(k)
             self._ids.add(v)
 
-    def _remove_ids(self, ids: list[str]):
+    def _remove_ids(self, ids: List[str]) -> None:
         for id_ in ids:
             try:
                 self._ids.remove(id_)
             except KeyError:
                 warnings.warn("Could not find id. Do you have duplicate ids in the model?")
 
-    def update_meta_info(self, component: str, meta_info: dict):
+    def update_meta_info(self, component: str, meta_info: dict) -> None:
         """Add meta info for any given model component.
 
         Parameters
@@ -133,7 +146,7 @@ class BaseModel(ABC):
             else:
                 setattr(self.meta_info[component], k1, d)
 
-    def _get_nonzero_meta_info(self, *, component: str):
+    def _get_nonzero_meta_info(self, *, component: str) -> Dict[str, Any]:
         """Get meta info of a component for all entries that are not None.
 
         Parameters
@@ -189,7 +202,7 @@ class BaseModel(ABC):
         sbml_model_fbc.setStrict(True)
         return sbml_model
 
-    def _create_sbml_units(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_units(self, *, sbml_model: libsbml.Model) -> None:
         """Create sbml units out of the meta_info.
 
         Parameters
@@ -205,7 +218,7 @@ class BaseModel(ABC):
             sbml_unit.setScale(unit["scale"])
             sbml_unit.setMultiplier(unit["multiplier"])
 
-    def _create_sbml_compartments(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_compartments(self, *, sbml_model: libsbml.Model) -> None:
         """Create the compartments for the sbml model.
 
         Since modelbase does not enforce any compartments, so far
@@ -225,10 +238,10 @@ class BaseModel(ABC):
             sbml_compartment.setUnits(compartment["units"])
 
     @abstractmethod
-    def _model_to_sbml(self):
+    def _model_to_sbml(self) -> None:
         """Define which methods shall be used for sbml export."""
 
-    def write_sbml_model(self, filename: str = None):
+    def write_sbml_model(self, filename: str = None) -> Optional[str]:
         """Write the model to an sbml file.
 
         Parameters
@@ -242,5 +255,7 @@ class BaseModel(ABC):
         doc = self._model_to_sbml()
         if filename is not None:
             libsbml.writeSBMLToFile(doc, filename)
+            return None
         else:
-            return libsbml.writeSBMLToString(doc)
+            model_str: str = libsbml.writeSBMLToString(doc)
+            return model_str
