@@ -3,34 +3,36 @@ from __future__ import annotations
 
 import warnings
 from dataclasses import dataclass, field
+from typing import Any, Dict, Iterable, List, Optional, Tuple
 
 import libsbml
 
+from .basemodel import BaseModel
 from .utils import convert_id_to_sbml, warning_on_one_line
 
-warnings.formatwarning = warning_on_one_line
+warnings.formatwarning = warning_on_one_line  # type: ignore
 
 
 @dataclass
 class Compound:
     """Meta-info container for compounds."""
 
-    common_name: str = None
-    compartment: str = "c"
-    unit: str = None
-    formula: str = None
-    charge: float = None
-    gibbs0: float = None
-    smiles: str = None
+    common_name: Optional[str] = None
+    compartment: Optional[str] = "c"
+    unit: Optional[str] = None
+    formula: Optional[str] = None
+    charge: Optional[float] = None
+    gibbs0: Optional[float] = None
+    smiles: Optional[str] = None
     database_links: dict = field(default_factory=dict)
     notes: dict = field(default_factory=dict)
 
 
-class CompoundMixin:
+class CompoundMixin(BaseModel):
     """Mixin for compound functionality."""
 
-    def __init__(self, compounds=None):
-        self.compounds = []
+    def __init__(self, compounds: List[str] = None) -> None:
+        self.compounds: List[str] = []
         if compounds is not None:
             self.add_compounds(compounds=compounds)
 
@@ -41,8 +43,8 @@ class CompoundMixin:
     def add_compound(
         self,
         compound: str,
-        **meta_info,
-    ):
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a compound to the model.
 
         Parameters
@@ -62,10 +64,13 @@ class CompoundMixin:
                 warnings.warn(f"Overwriting compound {compound}")
                 self.remove_compound(compound=compound)
             self.compounds.append(compound)
-            self.meta_info.setdefault("compounds", {}).setdefault(compound, Compound(**meta_info))
-            self._check_and_insert_ids([compound])
+            self.meta_info.setdefault("compounds", {}).setdefault(
+                compound,
+                Compound(**meta_info),  # type: ignore
+            )
+            self._check_and_insert_ids([compound], context="add_compound")
 
-    def add_compounds(self, compounds: list[str], meta_info=None):
+    def add_compounds(self, compounds: List[str], meta_info: Optional[Dict[str, Any]] = None) -> None:
         """Add multiple compounds to the model.
 
         See Also
@@ -80,7 +85,7 @@ class CompoundMixin:
                 info = {}
             self.add_compound(compound=compound, **info)
 
-    def update_compound_meta_info(self, compound: str, meta_info: dict):
+    def update_compound_meta_info(self, compound: str, meta_info: dict) -> None:
         """Update meta info of a compound.
 
         Parameters
@@ -93,7 +98,7 @@ class CompoundMixin:
         """
         self.update_meta_info(component="compounds", meta_info={compound: meta_info})
 
-    def remove_compound(self, compound: str):
+    def remove_compound(self, compound: str) -> None:
         self.compounds.remove(compound)
         self._remove_ids([compound])
         try:
@@ -101,18 +106,21 @@ class CompoundMixin:
         except KeyError:
             pass
 
-    def remove_compounds(self, compounds: list[str]):
+    def remove_compounds(self, compounds: Iterable[str]) -> None:
         for compound in compounds:
             self.remove_compound(compound=compound)
 
-    def get_compounds(self) -> list[str]:
+    def _get_all_compounds(self) -> List[str]:
+        return list(self.compounds)
+
+    def get_compounds(self) -> List[str]:
         return list(self.compounds)
 
     ##########################################################################
     # Source code functions
     ##########################################################################
 
-    def _generate_compounds_source_code(self, *, include_meta_info=True) -> str:
+    def _generate_compounds_source_code(self, *, include_meta_info: bool = True) -> str:
         """Generate modelbase source code for compounds.
 
         This is mainly used for the generate_model_source_code function.
@@ -137,7 +145,7 @@ class CompoundMixin:
     # SBML functions
     ##########################################################################
 
-    def _create_sbml_compounds(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_compounds(self, *, sbml_model: libsbml.Model) -> None:
         """Create the compounds for the sbml model.
 
         Parameters
