@@ -5,32 +5,33 @@ import json as json
 import pickle as pickle
 import warnings
 from dataclasses import dataclass, field
-from typing import Callable
+from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple
 
 import libsbml
 
+from .basemodel import BaseModel
 from .utils import convert_id_to_sbml, warning_on_one_line
 
-warnings.formatwarning = warning_on_one_line
+warnings.formatwarning = warning_on_one_line  # type: ignore
 
 
 @dataclass
 class Parameter:
     """Meta-info container for parameters."""
 
-    unit: str = None
-    annotation: str = None
+    unit: Optional[str] = None
+    annotation: Optional[str] = None
     database_links: dict = field(default_factory=dict)
     notes: dict = field(default_factory=dict)
 
 
-class ParameterMixin:
+class ParameterMixin(BaseModel):
     """Adding parameter functions."""
 
-    def __init__(self, parameters: dict = None):
-        self.parameters: dict[str, float] = {}
-        self.derived_parameters = {}
-        self._derived_from_parameters = set()
+    def __init__(self, parameters: dict = None) -> None:
+        self.parameters: Dict[str, float] = {}
+        self.derived_parameters: Dict[str, Dict[str, Any]] = {}
+        self._derived_from_parameters: set[str] = set()
 
         if parameters is not None:
             self.add_parameters(parameters=parameters)
@@ -40,7 +41,7 @@ class ParameterMixin:
     # Parameter functions
     ##########################################################################
 
-    def add_parameter(self, parameter_name: str, parameter_value: float, **meta_info):
+    def add_parameter(self, parameter_name: str, parameter_value: float, **meta_info: Dict[str, Any]) -> None:
         """Add a new parameter to the model.
 
         Parameters
@@ -49,11 +50,18 @@ class ParameterMixin:
             Meta info of the parameter. Allowed keys are
             {unit, database_links, notes}
         """
-        self._check_and_insert_ids([parameter_name])
+        self._check_and_insert_ids([parameter_name], context="add_parameter")
         self.add_and_update_parameter(parameter_name, parameter_value)
-        self.meta_info.setdefault("parameters", {}).setdefault(parameter_name, Parameter(**meta_info))
+        self.meta_info.setdefault("parameters", {}).setdefault(
+            parameter_name,
+            Parameter(**meta_info),  # type: ignore
+        )
 
-    def add_parameters(self, parameters: dict[str, float], meta_info=None):
+    def add_parameters(
+        self,
+        parameters: Dict[str, float],
+        meta_info: Optional[Dict[str, Any]] = None,
+    ) -> None:
         meta_info = {} if meta_info is None else meta_info
         for parameter_name, parameter_value in parameters.items():
             info = meta_info.get(parameter_name, {})
@@ -63,7 +71,7 @@ class ParameterMixin:
                 **info,
             )
 
-    def update_parameter(self, parameter_name, parameter_value):
+    def update_parameter(self, parameter_name: str, parameter_value: float) -> None:
         """Update a model parameter.
 
         Parameters
@@ -82,7 +90,7 @@ class ParameterMixin:
             warnings.warn(f"Key {parameter_name} is not in the model. Adding.")
         self.add_and_update_parameter(parameter_name, parameter_value)
 
-    def update_parameters(self, parameters):
+    def update_parameters(self, parameters: Dict[str, float]) -> None:
         """Update existing model parameters.
 
         Parameters
@@ -97,13 +105,18 @@ class ParameterMixin:
         for parameter_name, parameter_value in parameters.items():
             self.update_parameter(parameter_name=parameter_name, parameter_value=parameter_value)
 
-    def add_and_update_parameter(self, parameter_name: str, parameter_value: float, update_derived=True):
+    def add_and_update_parameter(
+        self,
+        parameter_name: str,
+        parameter_value: float,
+        update_derived: bool = True,
+    ) -> None:
         """Add a new or update an existing parameter."""
         self.parameters[parameter_name] = parameter_value
         if parameter_name in self._derived_from_parameters and update_derived:
             self._update_derived_parameters()
 
-    def add_and_update_parameters(self, parameters: dict[str, float]):
+    def add_and_update_parameters(self, parameters: Dict[str, float]) -> None:
         """Add new and updates existing model parameters.
 
         Parameters
@@ -118,7 +131,7 @@ class ParameterMixin:
         for parameter_name, parameter_value in parameters.items():
             self.add_and_update_parameter(parameter_name=parameter_name, parameter_value=parameter_value)
 
-    def update_parameter_meta_info(self, parameter: str, meta_info: dict):
+    def update_parameter_meta_info(self, parameter: str, meta_info: dict) -> None:
         """Update meta info of a parameter.
 
         Parameters
@@ -131,7 +144,7 @@ class ParameterMixin:
         """
         self.update_meta_info(component="parameters", meta_info={parameter: meta_info})
 
-    def remove_parameter(self, parameter_name: str):
+    def remove_parameter(self, parameter_name: str) -> None:
         """Remove a parameter from the model.
 
         Parameters
@@ -143,7 +156,7 @@ class ParameterMixin:
             del self.meta_info["parameters"][parameter_name]
         self._remove_ids([parameter_name])
 
-    def remove_parameters(self, parameter_names: list[str]):
+    def remove_parameters(self, parameter_names: Iterable[str]) -> None:
         """Remove parameters from the model.
 
         Parameters
@@ -158,7 +171,7 @@ class ParameterMixin:
         for parameter_name in parameter_names:
             self.remove_parameter(parameter_name=parameter_name)
 
-    def add_derived_parameter(self, parameter_name: str, function: Callable, parameters: list[str] = None):
+    def add_derived_parameter(self, parameter_name: str, function: Callable, parameters: List[str]) -> None:
         """Add a derived parameter.
 
         Derived parameters are calculated from other model parameters and dynamically updated
@@ -182,7 +195,7 @@ class ParameterMixin:
             parameter_value=function(*parameter_values),
         )
 
-    def remove_derived_parameter(self, parameter_name: str):
+    def remove_derived_parameter(self, parameter_name: str) -> None:
         """Remove a derived parameter from the model.
 
         Parameters
@@ -197,7 +210,7 @@ class ParameterMixin:
                 self._derived_from_parameters.remove(i)
         self.remove_parameter(parameter_name=parameter_name)
 
-    def _update_derived_parameters(self):
+    def _update_derived_parameters(self) -> None:
         """Update values of all derived parameters.
 
         This function is supposed to be run after a change to any parameter
@@ -206,16 +219,14 @@ class ParameterMixin:
         If there ever is a model with lots of derived parameters, it might
         make sense to make this a little less brute-force ;)
         """
-        for parameter_name, param_dict in self.derived_parameters.items():
+        for parameter_name, parameter in self.derived_parameters.items():
             self.add_and_update_parameter(
                 parameter_name=parameter_name,
-                parameter_value=param_dict["function"](
-                    *(self.parameters[i] for i in param_dict["parameters"])
-                ),
+                parameter_value=parameter["function"](*(self.parameters[i] for i in parameter["parameters"])),
                 update_derived=False,
             )
 
-    def store_parameters_to_file(self, filename: str, filetype="json"):
+    def store_parameters_to_file(self, filename: str, filetype: str = "json") -> None:
         """Store the parameters into a json or pickle file.
 
         Parameters
@@ -233,12 +244,12 @@ class ParameterMixin:
         elif filetype == "pickle":
             if not filename.endswith(".p"):
                 filename += ".p"
-            with open(filename, "wb") as f:
-                pickle.dump(self.parameters, f)
+            with open(filename, "wb") as f:  # type: ignore
+                pickle.dump(self.parameters, f)  # type: ignore
         else:
             raise ValueError("Can only save to json or pickle")
 
-    def load_parameters_from_file(self, filename: str, filetype="json"):
+    def load_parameters_from_file(self, filename: str, filetype: str = "json") -> None:
         """Load parameters from a json or pickle file.
 
         Parameters
@@ -252,12 +263,12 @@ class ParameterMixin:
             with open(filename, "r") as f:
                 self.add_and_update_parameters(parameters=json.load(f))
         elif filetype == "pickle":
-            with open(filename, "rb") as f:
-                self.add_and_update_parameters(parameters=pickle.load(f))
+            with open(filename, "rb") as f:  # type: ignore
+                self.add_and_update_parameters(parameters=pickle.load(f))  # type: ignore
         else:
             raise ValueError("Can only load from json or pickle")
 
-    def restore_initialization_parameters(self):
+    def restore_initialization_parameters(self) -> None:
         """Restore parameters to initialization parameters."""
         self.parameters = self.initialization_parameters.copy()
 
@@ -276,7 +287,7 @@ class ParameterMixin:
         """
         return float(self.parameters[parameter_name])
 
-    def get_parameters(self) -> dict[str, float]:
+    def get_parameters(self) -> Dict[str, float]:
         """Return all parameters.
 
         Returns
@@ -285,14 +296,14 @@ class ParameterMixin:
         """
         return dict(self.parameters)
 
-    def get_parameter_names(self) -> list[str]:
+    def get_parameter_names(self) -> List[str]:
         return list(self.parameters)
 
     ##########################################################################
     # Source code functions
     ##########################################################################
 
-    def _generate_parameters_source_code(self, *, include_meta_info=True) -> str:
+    def _generate_parameters_source_code(self, *, include_meta_info: bool = True) -> str:
         """Generate modelbase source code for parameters.
 
         This is mainly used for the generate_model_source_code function.
@@ -318,7 +329,7 @@ class ParameterMixin:
     # SBML functions
     ##########################################################################
 
-    def _create_sbml_parameters(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_parameters(self, *, sbml_model: libsbml.Model) -> None:
         """Create the parameters for the sbml model.
 
         Parameters
