@@ -2,22 +2,25 @@
 from __future__ import annotations
 
 import warnings
+from typing import Dict, List
 
+import libsbml
 import numpy as np
 import pandas as pd
-import libsbml
 
+from .basemodel import BaseModel
+from .compoundmixin import CompoundMixin
 from .utils import convert_id_to_sbml, warning_on_one_line
 
-warnings.formatwarning = warning_on_one_line
+warnings.formatwarning = warning_on_one_line  # type: ignore
 
 
-class StoichiometricMixin:
+class StoichiometricMixin(CompoundMixin, BaseModel):
     """Mixin for stoichiometries."""
 
-    def __init__(self, rate_stoichiometries: dict = None):
-        self.stoichiometries: dict[str, dict[str, float]] = {}
-        self.stoichiometries_by_compounds: dict[str, dict[str, float]] = {}
+    def __init__(self, rate_stoichiometries: dict = None) -> None:
+        self.stoichiometries: Dict[str, Dict[str, float]] = {}
+        self.stoichiometries_by_compounds: Dict[str, Dict[str, float]] = {}
         if rate_stoichiometries is not None:
             self.add_stoichiometries(rate_stoichiometries=rate_stoichiometries)
 
@@ -25,7 +28,7 @@ class StoichiometricMixin:
     # Stoichiometries
     ##########################################################################
 
-    def add_stoichiometry(self, rate_name: str, stoichiometry: dict[str, float]):
+    def add_stoichiometry(self, rate_name: str, stoichiometry: Dict[str, float]) -> None:
         """Add the stoichiometry of a rate to the model.
 
         Examples
@@ -43,7 +46,7 @@ class StoichiometricMixin:
         for compound, factor in stoichiometry.items():
             self.stoichiometries_by_compounds.setdefault(compound, {})[rate_name] = factor
 
-    def add_stoichiometry_by_compound(self, compound: str, stoichiometry: dict[str, float]):
+    def add_stoichiometry_by_compound(self, compound: str, stoichiometry: Dict[str, float]) -> None:
         """Add the stoichiometry of compound to the model.
 
         Examples
@@ -59,7 +62,7 @@ class StoichiometricMixin:
         for rate, factor in stoichiometry.items():
             self.stoichiometries.setdefault(rate, {})[compound] = factor
 
-    def add_stoichiometries(self, rate_stoichiometries: dict[str, dict[str, float]]):
+    def add_stoichiometries(self, rate_stoichiometries: Dict[str, Dict[str, float]]) -> None:
         """Add the stoichiometry of multiple rates.
 
         Examples
@@ -74,7 +77,7 @@ class StoichiometricMixin:
         for rate_name, stoichiometry in rate_stoichiometries.items():
             self.add_stoichiometry(rate_name=rate_name, stoichiometry=stoichiometry)
 
-    def update_stoichiometry(self, rate_name: str, stoichiometry: dict[str, float]):
+    def update_stoichiometry(self, rate_name: str, stoichiometry: Dict[str, float]) -> None:
         """Update stoichiometry of a rate"""
         if rate_name not in self.stoichiometries:
             warnings.warn("Rate not in stoichiometries!")
@@ -82,12 +85,12 @@ class StoichiometricMixin:
             self.remove_rate_stoichiometry(rate_name)
         self.add_stoichiometry(rate_name, stoichiometry)
 
-    def update_stoichiometries(self, stoichiometries: dict[str, dict[str, float]]):
+    def update_stoichiometries(self, stoichiometries: Dict[str, Dict[str, float]]) -> None:
         """Update stoichiometries of multiple rates"""
         for rate_name, stoichiometry in stoichiometries.items():
             self.update_stoichiometry(rate_name, stoichiometry)
 
-    def add_stoichiometries_by_compounds(self, compound_stoichiometries: dict[str, dict[str, float]]):
+    def add_stoichiometries_by_compounds(self, compound_stoichiometries: Dict[str, Dict[str, float]]) -> None:
         """Add the stoichiometry of multiple compounds to the model.
 
         Examples
@@ -102,7 +105,7 @@ class StoichiometricMixin:
         for compound, stoichiometry in compound_stoichiometries.items():
             self.add_stoichiometry_by_compound(compound=compound, stoichiometry=stoichiometry)
 
-    def remove_rate_stoichiometry(self, rate_name: str):
+    def remove_rate_stoichiometry(self, rate_name: str) -> None:
         """Remove a rate stoichiometry from the model."""
         compounds = self.stoichiometries.pop(rate_name)
         for compound in compounds:
@@ -110,12 +113,12 @@ class StoichiometricMixin:
             if not bool(self.stoichiometries_by_compounds[compound]):
                 del self.stoichiometries_by_compounds[compound]
 
-    def remove_rate_stoichiometries(self, rate_names: list[str]):
+    def remove_rate_stoichiometries(self, rate_names: List[str]) -> None:
         """Remove multiple rate stoichiometries from the model."""
         for rate_name in rate_names:
             self.remove_rate_stoichiometry(rate_name=rate_name)
 
-    def remove_compound_stoichiometry(self, compound: str):
+    def remove_compound_stoichiometry(self, compound: str) -> None:
         """Remove stoichiometry of a compound."""
         rates = self.stoichiometries_by_compounds.pop(compound)
         for rate in rates:
@@ -123,24 +126,24 @@ class StoichiometricMixin:
             if not bool(self.stoichiometries[rate]):
                 del self.stoichiometries[rate]
 
-    def remove_compound_stoichiometries(self, compounds: list[str]):
+    def remove_compound_stoichiometries(self, compounds: List[str]) -> None:
         """Remove stoichiometry of multiple compounds."""
         for compound in compounds:
             self.remove_compound_stoichiometry(compound=compound)
 
-    def get_rate_stoichiometry(self, rate_name: str) -> dict[str, float]:
+    def get_rate_stoichiometry(self, rate_name: str) -> Dict[str, float]:
         """Get stoichiometry of a rate."""
         return dict(self.stoichiometries[rate_name])
 
-    def get_compound_stoichiometry(self, compound: str) -> dict[str, float]:
+    def get_compound_stoichiometry(self, compound: str) -> Dict[str, float]:
         """Get stoichiometry of a compound."""
         return dict(self.stoichiometries_by_compounds[compound])
 
-    def get_stoichiometries(self) -> dict[str, dict[str, float]]:
+    def get_stoichiometries(self) -> Dict[str, Dict[str, float]]:
         """Return stoichiometries ordered by reactions."""
         return dict(self.stoichiometries)
 
-    def get_stoichiometries_by_compounds(self) -> dict[str, dict[str, float]]:
+    def get_stoichiometries_by_compounds(self) -> Dict[str, Dict[str, float]]:
         """Return stoichiometries ordered by compounds."""
         return dict(self.stoichiometries_by_compounds)
 
@@ -179,7 +182,7 @@ class StoichiometricMixin:
     # SBML functions
     ##########################################################################
 
-    def _create_sbml_stoichiometries(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_stoichiometries(self, *, sbml_model: libsbml.Model) -> None:
         """Create the reactions for the sbml model.
 
         Parameters
