@@ -3,28 +3,29 @@ from __future__ import annotations
 
 import itertools as it
 import warnings
-from collections import defaultdict as _defaultdict
-from typing import Union, Callable
+from collections import defaultdict
+from dataclasses import dataclass
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
 
 import libsbml
 import numpy as np
 import pandas as pd
 
 from ...core.compoundmixin import Compound
-from ...core.ratemixin import Rate
+from ...core.ratemixin import RateMeta
 from ...core.utils import convert_id_to_sbml, warning_on_one_line
 from .model import Model
 
-warnings.formatwarning = warning_on_one_line
+warnings.formatwarning = warning_on_one_line  # type: ignore
 
 
-def total_concentration(*args):
+def total_concentration(*args: float) -> List[np.ndarray]:
     """Return concentration of all isotopomers.
 
     Algebraic module function to keep track of the total
     concentration of a compound (so sum of its isotopomers).
     """
-    return [np.sum(args, axis=0)]
+    return [np.sum(args, axis=0)]  # type: ignore
 
 
 class LabelModel(Model):
@@ -33,15 +34,15 @@ class LabelModel(Model):
     def __init__(
         self,
         parameters: dict = None,
-        compounds: dict = None,
+        compounds: List[str] = None,
         algebraic_modules: dict = None,
-        rate_stoichiometries: dict = None,
+        rate_stoichiometries: Dict[str, Dict[str, int]] = None,
         rates: dict = None,
         meta_info: dict = None,
     ):
-        self.label_compounds = {}
-        self.nonlabel_compounds = []
-        self.base_reactions = {}
+        self.label_compounds: Dict[str, Dict[str, Any]] = {}
+        self.nonlabel_compounds: List[str] = []
+        self.base_reactions: Dict[str, Dict[str, Any]] = {}
         super().__init__(
             parameters=parameters,
             compounds=compounds,
@@ -52,7 +53,7 @@ class LabelModel(Model):
         )
 
     @staticmethod
-    def _generate_binary_labels(*, base_name: str, num_labels: int) -> list[str]:
+    def _generate_binary_labels(*, base_name: str, num_labels: int) -> List[str]:
         """Create binary label string.
 
         Returns
@@ -76,7 +77,12 @@ class LabelModel(Model):
         else:
             return [base_name]
 
-    def add_compound(self, compound: str, is_isotopomer=False, **meta_info):
+    def add_compound(  # type: ignore
+        self,
+        compound: str,
+        is_isotopomer: bool = False,
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a single compound to the model.
 
         Parameters
@@ -94,16 +100,26 @@ class LabelModel(Model):
         if not is_isotopomer:
             self.nonlabel_compounds.append(compound)
 
-    def _add_base_compound(self, *, base_compound: str, num_labels: int, label_names: list[str], **meta_info):
+    def _add_base_compound(
+        self,
+        *,
+        base_compound: str,
+        num_labels: int,
+        label_names: List[str],
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a base compound of label isotopomer."""
         self.label_compounds[base_compound] = {
             "num_labels": num_labels,
             "isotopomers": label_names,
         }
-        self._check_and_insert_ids([base_compound])
-        self.meta_info.setdefault("compounds", {}).setdefault(base_compound, Compound(**meta_info))
+        self._check_and_insert_ids([base_compound], context="base_compound")
+        self.meta_info.setdefault("compounds", {}).setdefault(
+            base_compound,
+            Compound(**meta_info),  # type: ignore
+        )
 
-    def _add_isotopomers(self, *, base_compound: str, label_names: list[str]):
+    def _add_isotopomers(self, *, base_compound: str, label_names: List[str]) -> None:
         # Add all labelled compounds
         for compound in label_names:
             self.add_compound(compound=compound, is_isotopomer=True)
@@ -119,7 +135,7 @@ class LabelModel(Model):
             parameters=None,
         )
 
-    def add_label_compound(self, compound: str, num_labels: int, **meta_info):
+    def add_label_compound(self, compound: str, num_labels: int, **meta_info: Dict[str, Any]) -> None:
         """Create all label isotopomers and add them as compounds.
 
         Also create an algebraic module that tracks the total
@@ -141,18 +157,26 @@ class LabelModel(Model):
             warnings.warn(f"Overwriting compound {compound}")
             self.remove_label_compound(compound=compound)
         if num_labels == 0:
-            self.add_compound(compound=compound, is_isotopomer=False, **meta_info)
+            self.add_compound(
+                compound=compound,
+                is_isotopomer=False,
+                **meta_info,  # type: ignore
+            )
         else:
             label_names = self._generate_binary_labels(base_name=compound, num_labels=num_labels)
             self._add_base_compound(
                 base_compound=compound,
                 num_labels=num_labels,
                 label_names=label_names,
-                **meta_info,
+                **meta_info,  # type: ignore
             )
             self._add_isotopomers(base_compound=compound, label_names=label_names)
 
-    def add_label_compounds(self, compounds: dict[str, int], meta_info=None):
+    def add_label_compounds(
+        self,
+        compounds: Dict[str, int],
+        meta_info: Optional[Dict[str, Any]] = None,
+    ) -> None:
         """Add multiple label-containing compounds to the model.
 
         Parameters
@@ -174,7 +198,7 @@ class LabelModel(Model):
             info = meta_info.get(compound, {})
             self.add_label_compound(compound=compound, num_labels=num_labels, **info)
 
-    def remove_compound(self, compound: str, is_isotopomer=False):
+    def remove_compound(self, compound: str, is_isotopomer: bool = False) -> None:
         """Remove a compound from the model.
 
         Parameters
@@ -188,19 +212,19 @@ class LabelModel(Model):
         if not is_isotopomer:
             self.nonlabel_compounds.remove(compound)
 
-    def remove_label_compound(self, compound: str):
+    def remove_label_compound(self, compound: str) -> None:
         """Remove a label compound from the model."""
         label_compound = self.label_compounds.pop(compound)
         self._remove_ids([compound])
         for key in label_compound["isotopomers"]:
             self.remove_compound(compound=key, is_isotopomer=True)
 
-    def remove_label_compounds(self, compounds: list[str]):
+    def remove_label_compounds(self, compounds: List[str]) -> None:
         """Remove label compounds."""
         for compound in compounds:
             self.remove_label_compound(compound=compound)
 
-    def get_base_compounds(self) -> list[str]:
+    def get_base_compounds(self) -> List[str]:
         """Get all base compounds and non-label compounds."""
         return sorted(list(self.label_compounds) + self.nonlabel_compounds)
 
@@ -208,11 +232,11 @@ class LabelModel(Model):
         """Get the number of possible labels positions of a compound."""
         return int(self.label_compounds[compound]["num_labels"])
 
-    def get_compound_isotopomers(self, compound: str) -> list[str]:
+    def get_compound_isotopomers(self, compound: str) -> List[str]:
         """Get all isotopomers of a compound."""
         return list(self.label_compounds[compound]["isotopomers"])
 
-    def get_compound_isotopomers_with_n_labels(self, compound: str, n_labels: int) -> list[str]:
+    def get_compound_isotopomers_with_n_labels(self, compound: str, n_labels: int) -> List[str]:
         """Get all isotopomers of a compound, that have excactly n labels."""
         label_positions = self.label_compounds[compound]["num_labels"]
         label_patterns = [
@@ -222,8 +246,8 @@ class LabelModel(Model):
         return [f"{compound}__{''.join(i)}" for i in label_patterns]
 
     def get_compound_isotopomer_with_label_position(
-        self, base_compound: str, label_position: Union[int, list[int]]
-    ) -> list[str]:
+        self, base_compound: str, label_position: Union[int, List[int]]
+    ) -> str:
         """Get compound isotopomer with a given label position.
 
         Examples
@@ -241,7 +265,7 @@ class LabelModel(Model):
         )
 
     @staticmethod
-    def _split_label_string(label: str, *, labels_per_compound: list[str]) -> list[str]:
+    def _split_label_string(label: str, *, labels_per_compound: List[int]) -> List[str]:
         """Split label string according to labels given in label list.
 
         The labels in the label list correspond to the number of
@@ -275,12 +299,12 @@ class LabelModel(Model):
         return split_labels
 
     @staticmethod
-    def _map_substrates_to_products(*, rate_suffix: str, labelmap: list[int]) -> list[str]:
+    def _map_substrates_to_products(*, rate_suffix: str, labelmap: List[int]) -> str:
         """Map the rate_suffix to products using the labelmap."""
         return "".join([rate_suffix[i] for i in labelmap])
 
     @staticmethod
-    def _unpack_stoichiometries(*, stoichiometries: dict[str, int]) -> tuple[list[str]]:
+    def _unpack_stoichiometries(*, stoichiometries: Dict[str, int]) -> Tuple[List[str], List[str]]:
         """Split stoichiometries into substrates and products.
 
         Parameters
@@ -301,7 +325,7 @@ class LabelModel(Model):
                 products.extend([k] * v)
         return substrates, products
 
-    def _get_labels_per_compound(self, *, compounds: list[str]) -> list[int]:
+    def _get_labels_per_compound(self, *, compounds: List[str]) -> List[int]:
         """Get labels per compound.
 
         This is used for _split_label string. Adds 0 for non-label compounds,
@@ -317,9 +341,9 @@ class LabelModel(Model):
         return labels_per_compound
 
     @staticmethod
-    def _repack_stoichiometries(*, new_substrates: list[str], new_products: list[str]) -> dict[str, int]:
+    def _repack_stoichiometries(*, new_substrates: List[str], new_products: List[str]) -> Dict[str, float]:
         """Pack substrates and products into stoichiometric dict."""
-        new_stoichiometries = _defaultdict(int)
+        new_stoichiometries: defaultdict[str, int] = defaultdict(int)
         for arg in new_substrates:
             new_stoichiometries[arg] -= 1
         for arg in new_products:
@@ -327,7 +351,7 @@ class LabelModel(Model):
         return dict(new_stoichiometries)
 
     @staticmethod
-    def _assign_compound_labels(*, base_compounds: list[str], label_suffixes: list[str]) -> list[str]:
+    def _assign_compound_labels(*, base_compounds: List[str], label_suffixes: List[str]) -> List[str]:
         """Assign the correct suffixes."""
         new_compounds = []
         for i, compound in enumerate(base_compounds):
@@ -337,15 +361,15 @@ class LabelModel(Model):
                 new_compounds.append(compound)
         return new_compounds
 
-    def add_algebraic_module(
+    def add_algebraic_module(  # type: ignore
         self,
         module_name: str,
         function: Callable,
-        compounds: list[str] = None,
-        derived_compounds: list[str] = None,
-        modifiers: list[str] = None,
-        parameters: list[str] = None,
-        args: list[str] = None,
+        compounds: List[str] = None,
+        derived_compounds: List[str] = None,
+        modifiers: List[str] = None,
+        parameters: List[str] = None,
+        args: List[str] = None,
     ):
         """Add an algebraic module to the model.
 
@@ -381,7 +405,7 @@ class LabelModel(Model):
 
         Examples
         --------
-        >>> def rapid_equilibrium(substrate, k_eq):
+        >>> def rapid_equilibrium(substrate, k_eq)-> None:
         >>>    x = substrate / (1 + k_eq)
         >>>    y = substrate * k_eq / (1 + k_eq)
         >>>    return x, y
@@ -414,40 +438,39 @@ class LabelModel(Model):
         rate_name: str,
         total_product_labels: int,
         total_substrate_labels: int,
-        external_labels: list[str],
+        external_labels: Optional[List[int]],
     ) -> str:
         n_external_labels = total_product_labels - total_substrate_labels
         if n_external_labels > 0:
-            if not external_labels:
+            if external_labels is None:
                 warnings.warn(f"Added external labels for reaction {rate_name}")
                 external_label_string = ["1"] * n_external_labels
             else:
                 external_label_string = ["0"] * n_external_labels
                 for label_pos in external_labels:
                     external_label_string[label_pos] = "1"
-            external_labels = "".join(external_label_string)
+            return "".join(external_label_string)
         else:
-            external_labels = ""
-        return external_labels
+            return ""
 
     def _add_base_reaction(
         self,
         *,
         rate_name: str,
         function: Callable,
-        stoichiometry: dict[str, float],
-        base_substrates: list[str],
-        base_products: list[str],
-        labelmap: list[int],
-        external_labels: str,
-        modifiers: list[str],
-        parameters: list[str],
+        stoichiometry: Dict[str, int],
+        base_substrates: List[str],
+        base_products: List[str],
+        labelmap: List[int],
+        external_labels: Optional[str],
+        modifiers: Optional[List[str]],
+        parameters: Optional[List[str]],
         reversible: bool,
-        # dynamic_variables: list[str] = None,
-        # args: list[str] = None,
-        variants: list[str],
-        **meta_info,
-    ):
+        # dynamic_variables: List[str] = None,
+        # args: List[str] = None,
+        variants: List[str],
+        **meta_info: Dict[str, Any],
+    ) -> None:
         # The check isn't as easy.
         # Don't compare to the compounds, probably against base compounds, non-isotopomers and derived compounds?
         # if parameters is None:
@@ -474,26 +497,31 @@ class LabelModel(Model):
             "reversible": reversible,
             "variants": variants,
         }
-        self.meta_info.setdefault("rates", {}).setdefault(rate_name, Rate(**meta_info))
+        self.meta_info.setdefault("rates", {}).setdefault(
+            rate_name,
+            RateMeta(
+                **meta_info,  # type: ignore
+            ),
+        )
 
     def _create_isotopomer_reactions(
         self,
         *,
         rate_name: str,
         function: Callable,
-        labelmap: list[int],
-        modifiers: list[str],
-        parameters: list[str],
+        labelmap: List[int],
+        modifiers: Optional[List[str]],
+        parameters: Optional[List[str]],
         reversible: bool,
-        # dynamic_variables: list[str],
-        # args: list[str],
+        # dynamic_variables: List[str],
+        # args: List[str],
         external_labels: str,
         total_substrate_labels: int,
-        base_substrates: list[str],
-        base_products: list[str],
-        labels_per_substrate: list[str],
-        labels_per_product: list[str],
-    ):
+        base_substrates: List[str],
+        base_products: List[str],
+        labels_per_substrate: List[int],
+        labels_per_product: List[int],
+    ) -> List[str]:
         variants = []
         for rate_suffix in ("".join(i) for i in it.product(("0", "1"), repeat=total_substrate_labels)):
             rate_suffix += external_labels
@@ -535,16 +563,16 @@ class LabelModel(Model):
         self,
         rate_name: str,
         function: Callable,
-        stoichiometry: dict[str, float],
-        labelmap: list[int],
-        external_labels: Union[int, list[int]] = None,
-        modifiers: list[str] = None,
-        parameters: list[str] = None,
+        stoichiometry: Dict[str, int],
+        labelmap: List[int],
+        external_labels: Optional[List[int]] = None,
+        modifiers: Optional[List[str]] = None,
+        parameters: Optional[List[str]] = None,
         reversible: bool = False,
-        # dynamic_variables: list[str] = None,
-        # args: list[str] = None,
-        **meta_info,
-    ):
+        # dynamic_variables: List[str] = None,
+        # args: List[str] = None,
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a labelmap reaction.
 
         Parameters
@@ -600,7 +628,7 @@ class LabelModel(Model):
         if len(labelmap) - total_substrate_labels < 0:
             raise ValueError(f"Labelmap 'missing' {abs(len(labelmap) - total_substrate_labels)} label(s)")
 
-        external_labels = self._get_external_labels(
+        external_label_str = self._get_external_labels(
             rate_name=rate_name,
             total_product_labels=total_product_labels,
             total_substrate_labels=total_substrate_labels,
@@ -616,7 +644,7 @@ class LabelModel(Model):
             reversible=reversible,
             # dynamic_variables=dynamic_variables,
             # args=args,
-            external_labels=external_labels,
+            external_labels=external_label_str,
             total_substrate_labels=total_substrate_labels,
             base_substrates=base_substrates,
             base_products=base_products,
@@ -631,7 +659,7 @@ class LabelModel(Model):
             base_products=base_products,
             stoichiometry=stoichiometry,
             labelmap=labelmap,
-            external_labels=external_labels,
+            external_labels=external_label_str,
             modifiers=modifiers,
             parameters=parameters,
             reversible=reversible,
@@ -645,14 +673,14 @@ class LabelModel(Model):
         self,
         rate_name: str,
         function: Callable = None,
-        stoichiometry: dict[str, int] = None,
-        labelmap: list[int] = None,
-        modifiers: list[str] = None,
-        parameters: list[str] = None,
+        stoichiometry: Dict[str, int] = None,
+        labelmap: List[int] = None,
+        modifiers: List[str] = None,
+        parameters: List[str] = None,
         reversible: bool = None,
-        # dynamic_variables: list[str] = None,
-        # args: list[str] = None,
-    ):
+        # dynamic_variables: List[str] = None,
+        # args: List[str] = None,
+    ) -> None:
         """Update an existing labelmap reaction.
 
         Parameters
@@ -700,7 +728,7 @@ class LabelModel(Model):
             # args=args,
         )
 
-    def remove_labelmap_reaction(self, rate_name: str):
+    def remove_labelmap_reaction(self, rate_name: str) -> None:
         """Remove all variants of a base reaction.
 
         Parameters
@@ -714,7 +742,7 @@ class LabelModel(Model):
             if rate.startswith(rate_name):
                 self.remove_reaction(rate_name=rate)
 
-    def remove_labelmap_reactions(self, rate_names: list[str]):
+    def remove_labelmap_reactions(self, rate_names: List[str]) -> None:
         """Remove all variants of a multiple labelmap reactions.
 
         Parameters
@@ -729,8 +757,10 @@ class LabelModel(Model):
             self.remove_labelmap_reaction(rate_name=rate_name)
 
     def generate_y0(
-        self, base_y0: dict[str, float], label_positions: dict[str, Union[int, list[int]]] = None
-    ) -> dict[str, float]:
+        self,
+        base_y0: Union[List[float], Dict[str, float]],
+        label_positions: Dict[str, Union[int, List[int]]] = None,
+    ) -> Dict[str, float]:
         """Generate y0 for all isotopomers given a base y0.
 
         Examples
@@ -765,9 +795,9 @@ class LabelModel(Model):
     def get_total_fluxes(
         self,
         rate_base_name: str,
-        y: Union[dict[str, float], dict[str, list[float]]],
-        t: Union[float, list[float]] = 0,
-    ):
+        y: Union[Dict[str, float], Dict[str, List[float]]],
+        t: Union[float, List[float]] = 0,
+    ) -> pd.DataFrame:
         """Get total fluxes of a base rate.
 
         Parameters
@@ -783,34 +813,35 @@ class LabelModel(Model):
         rates = [i for i in self.rates if i.startswith(rate_base_name + "__")]
         return self.get_fluxes_df(y=y, t=t)[rates].sum(axis=1).values
 
-    def _create_label_scope_seed(self, *, initial_labels: dict[str, int]) -> dict[str, int]:
+    def _create_label_scope_seed(
+        self, *, initial_labels: Union[Dict[str, int], Dict[str, List[int]]]
+    ) -> Dict[str, int]:
         """Create initial label scope seed."""
-        initial_labels = [
-            self.get_compound_isotopomer_with_label_position(
-                base_compound=base_compound, label_position=label_position
-            )
-            for base_compound, label_position in initial_labels.items()
-        ]
 
         # initialise all compounds with 0 (no label)
         labelled_compounds = {compound: 0 for compound in self.get_compounds()}
 
         # Set all unlabelled compounds to 1
-        for compound, cpd_dict in self.label_compounds.items():
-            num_labels = cpd_dict["num_labels"]
-            labelled_compounds[f"{compound}__{'0' * num_labels}"] = 1
+        for name, compound in self.label_compounds.items():
+            num_labels = compound["num_labels"]
+            labelled_compounds[f"{name}__{'0' * num_labels}"] = 1
         # Also set all non-label compounds to 1
-        for compound in self.nonlabel_compounds:
-            labelled_compounds[compound] = 1
+        for name in self.nonlabel_compounds:
+            labelled_compounds[name] = 1
         # Set initial label
-        for i in initial_labels:
+        for i in [
+            self.get_compound_isotopomer_with_label_position(
+                base_compound=base_compound, label_position=label_position
+            )
+            for base_compound, label_position in initial_labels.items()
+        ]:
             labelled_compounds[i] = 1
         return labelled_compounds
 
     def get_label_scope(
         self,
-        initial_labels: Union[dict[str, int], dict[str, list[int]]],
-    ) -> dict[int, set[str]]:
+        initial_labels: Union[Dict[str, int], Dict[str, List[int]]],
+    ) -> Dict[int, set[str]]:
         """Return all label positions that can be reached step by step.
 
         Parameters
@@ -891,17 +922,17 @@ class LabelModel(Model):
 
         for module_name, module in self.algebraic_modules.items():
             if not module_name.endswith("__total"):
-                module = module.copy()
-                module["compounds"] = [
+                module_ = dict(module)
+                module_["compounds"] = [
                     self._map_label_compound_to_compound(compound=i) for i in module["compounds"]
                 ]
-                module["derived_compounds"] = [
+                module_["derived_compounds"] = [
                     self._map_label_compound_to_compound(compound=i) for i in module["derived_compounds"]
                 ]
-                module["modifiers"] = [
+                module_["modifiers"] = [
                     self._map_label_compound_to_compound(compound=i) for i in module["modifiers"]
                 ]
-                m.add_algebraic_module(module_name=module_name, **module)
+                m.add_algebraic_module(module_name=module_name, **module_)
 
         for rate_name, reaction in self.base_reactions.items():
             reaction = reaction.copy()
@@ -931,7 +962,7 @@ class LabelModel(Model):
         compartment: str,
         charge: int,
         formula: str,
-    ):
+    ) -> None:
         cpd = sbml_model.createSpecies()
         cpd.setId(convert_id_to_sbml(id_=compound_id, prefix="CPD"))
         if common_name is not None:
@@ -947,7 +978,7 @@ class LabelModel(Model):
         if formula is not None:
             cpd_fbc.setChemicalFormula(formula)
 
-    def _create_sbml_compounds(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_compounds(self, *, sbml_model: libsbml.Model) -> None:
         """Create the compounds for the sbml model.
 
         Parameters
@@ -986,7 +1017,7 @@ class LabelModel(Model):
                 formula=formula,
             )
 
-    def _create_sbml_reactions(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_reactions(self, *, sbml_model: libsbml.Model) -> None:
         """Create the reactions for the sbml model.
 
         Parameters
