@@ -1,12 +1,9 @@
-"""Main ODE model module.
-
-Description of the module
-"""
 from __future__ import annotations
 
+import itertools as it
 import subprocess
 import warnings
-from typing import Union, Iterable, Callable
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
 
 import libsbml
 import numpy as np
@@ -21,22 +18,16 @@ from ...core import (
     StoichiometricMixin,
 )
 from ...core.utils import convert_id_to_sbml
+from . import _AbstractRateModel
 
 
-class Model(
-    RateMixin,
-    StoichiometricMixin,
-    AlgebraicMixin,
-    ParameterMixin,
-    CompoundMixin,
-    BaseModel,
-):
+class Model(_AbstractRateModel):
     """The main class for modeling. Provides model construction and inspection tools."""
 
     def __init__(
         self,
         parameters: dict = None,
-        compounds: dict = None,
+        compounds: List[str] = None,
         algebraic_modules: dict = None,
         rate_stoichiometries: dict = None,
         rates: dict = None,
@@ -51,7 +42,7 @@ class Model(
         RateMixin.__init__(self, rates=rates, functions=functions)
         self.meta_info["model"].sbo = "SBO:0000062"  # continuous framework
 
-    def __str__(self):
+    def __str__(self) -> str:
         """Give a string representation.
 
         Returns
@@ -64,7 +55,9 @@ class Model(
             + f"\n    {len(self.get_stoichiometries())} Reactions"
         )
 
-    def _element_difference(self, other: Model, attribute: str):
+    def _element_difference(
+        self, other: Model, attribute: str
+    ) -> Optional[Union[List[str], Dict[str, List[str]]]]:
         self_collection = getattr(self, attribute)
         other_collection = getattr(other, attribute)
         difference = sorted(set(other_collection).difference(self_collection))
@@ -74,10 +67,12 @@ class Model(
             return difference
         return {k: other_collection[k] for k in difference}
 
-    def _element_intersection(self, other: Model, attribute: str):
+    def _element_intersection(
+        self, other: Model, attribute: str
+    ) -> Optional[Union[List[str], Dict[str, List[str]]]]:
         self_collection = getattr(self, attribute)
         other_collection = getattr(other, attribute)
-        intersection: list[str] = sorted(set(self_collection).intersection(other_collection))
+        intersection: List[str] = sorted(set(self_collection).intersection(other_collection))
         if not intersection:
             return None
         if attribute == "compounds":
@@ -85,45 +80,48 @@ class Model(
         return {k: other_collection[k] for k in intersection}
 
     def __add__(self, other: Model) -> Model:
-        m = self.copy()
-        m.add(**{"compounds": self._element_difference(other, "compounds")})
+        m: Model = self.copy()
+        m.add(**{"compounds": self._element_difference(other, "compounds")})  # type: ignore
+        m.add(**{"parameters": self._element_difference(other, "parameters")})  # type: ignore
+        m.update(**{"parameters": self._element_intersection(other, "parameters")})  # type: ignore
 
         for name, module in other.algebraic_modules.items():
             if name in self.algebraic_modules:
-                m.update_algebraic_module(name, **module)
+                m.update_algebraic_module(name, **dict(module))
             else:
-                m.add_algebraic_module(name, **module)
+                m.add_algebraic_module(name, **dict(module))
 
         for attribute in [
-            "parameters",
             "rates",
             "stoichiometries",
             "functions",
         ]:
-            m.add(**{attribute: self._element_difference(other, attribute)})
-            m.update(**{attribute: self._element_intersection(other, attribute)})
+            m.add(**{attribute: self._element_difference(other, attribute)})  # type: ignore
+            m.update(**{attribute: self._element_intersection(other, attribute)})  # type: ignore
         return m
 
     def __iadd__(self, other: Model) -> Model:
-        self.add(**{"compounds": self._element_difference(other, "compounds")})
+        self.add(**{"compounds": self._element_difference(other, "compounds")})  # type: ignore
+        self.add(**{"parameters": self._element_difference(other, "parameters")})  # type: ignore
+        self.update(**{"parameters": self._element_intersection(other, "parameters")})  # type: ignore
+
         for name, module in other.algebraic_modules.items():
             if name in self.algebraic_modules:
-                self.update_algebraic_module(name, **module)
+                self.update_algebraic_module(name, **dict(module))
             else:
-                self.add_algebraic_module(name, **module)
+                self.add_algebraic_module(name, **dict(module))
 
         for attribute in [
-            "parameters",
             "rates",
             "stoichiometries",
             "functions",
         ]:
-            self.add(**{attribute: self._element_difference(other, attribute)})
-            self.update(**{attribute: self._element_intersection(other, attribute)})
+            self.add(**{attribute: self._element_difference(other, attribute)})  # type: ignore
+            self.update(**{attribute: self._element_intersection(other, attribute)})  # type: ignore
         return self
 
     def __sub__(self, other: Model) -> Model:
-        m = self.copy()
+        m: Model = self.copy()
         for attribute in [
             "compounds",
             "parameters",
@@ -132,7 +130,7 @@ class Model(
             "stoichiometries",
             "functions",
         ]:
-            m.remove(**{attribute: self._element_intersection(other, attribute)})
+            m.remove(**{attribute: self._element_intersection(other, attribute)})  # type: ignore
         return m
 
     def __isub__(self, other: Model) -> Model:
@@ -144,18 +142,18 @@ class Model(
             "stoichiometries",
             "functions",
         ]:
-            self.remove(**{attribute: self._element_intersection(other, attribute)})
+            self.remove(**{attribute: self._element_intersection(other, attribute)})  # type: ignore
         return self
 
     def add(
         self,
         compounds: list = None,
-        parameters: dict[str, float] = None,
-        algebraic_modules: dict[str, dict] = None,
-        rates: dict[str, dict] = None,
-        stoichiometries: dict[str, dict] = None,
-        functions: dict[str, callable] = None,
-    ):
+        parameters: Dict[str, float] = None,
+        algebraic_modules: Dict[str, dict] = None,
+        rates: Dict[str, dict] = None,
+        stoichiometries: Dict[str, dict] = None,
+        functions: Dict[str, Callable] = None,
+    ) -> None:
         if compounds is not None:
             self.add_compounds(compounds)
         if parameters is not None:
@@ -171,12 +169,12 @@ class Model(
 
     def update(
         self,
-        parameters: dict[str, float] = None,
-        algebraic_modules: dict[str, dict] = None,
-        rates: dict[str, dict] = None,
-        stoichiometries: dict[str, dict] = None,
-        functions: dict[str, callable] = None,
-    ):
+        parameters: Dict[str, float] = None,
+        algebraic_modules: Dict[str, dict] = None,
+        rates: Dict[str, dict] = None,
+        stoichiometries: Dict[str, dict] = None,
+        functions: Dict[str, Callable] = None,
+    ) -> None:
         if parameters is not None:
             self.update_parameters(parameters)
         if algebraic_modules is not None:
@@ -190,13 +188,13 @@ class Model(
 
     def remove(
         self,
-        compounds: list[str] = None,
-        parameters: list[str] = None,
-        algebraic_modules: list[str] = None,
-        rates: list[str] = None,
-        stoichiometries: list[str] = None,
-        functions: list[str] = None,
-    ):
+        compounds: List[str] = None,
+        parameters: List[str] = None,
+        algebraic_modules: List[str] = None,
+        rates: List[str] = None,
+        stoichiometries: List[str] = None,
+        functions: List[str] = None,
+    ) -> None:
         if compounds is not None:
             self.remove_compounds(compounds)
         if parameters is not None:
@@ -217,16 +215,16 @@ class Model(
     def add_reaction(
         self,
         rate_name: str,
-        function: Callable,
-        stoichiometry: dict[str, int],
-        modifiers: list[str] = None,
-        parameters: list[str] = None,
-        dynamic_variables: list[str] = None,
-        args: list[str] = None,
+        function: Callable[..., float],
+        stoichiometry: Dict[str, float],
+        modifiers: List[str] = None,
+        parameters: List[str] = None,
+        dynamic_variables: List[str] = None,
+        args: List[str] = None,
         reversible: bool = False,
         check_consistency: bool = True,
-        **meta_info,
-    ):
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a reaction to the model.
 
         Shortcut for add_rate and add stoichiometry functions.
@@ -271,29 +269,45 @@ class Model(
         )
         self.add_stoichiometry(rate_name=rate_name, stoichiometry=stoichiometry)
 
+    def add_reaction_from_args(
+        self,
+        rate_name: str,
+        function: Callable[..., float],
+        stoichiometry: Dict[str, float],
+        args: List[str],
+        reversible: bool = False,
+    ) -> None:
+        modifiers = []
+        parameters = []
+        for i in args:
+            if i in self.parameters:
+                parameters.append(i)
+            elif i not in stoichiometry:
+                modifiers.append(i)
+        self.add_reaction(
+            rate_name=rate_name,
+            function=function,
+            stoichiometry=stoichiometry,
+            modifiers=modifiers,
+            parameters=parameters,
+            args=args,
+            reversible=reversible,
+        )
+
     def update_reaction(
         self,
         rate_name: str,
         function: Callable = None,
-        stoichiometry: dict[str, float] = None,
-        modifiers: list[str] = None,
-        parameters: list[str] = None,
-        dynamic_variables: list[str] = None,
-        args: list[str] = None,
+        stoichiometry: Dict[str, float] = None,
+        modifiers: List[str] = None,
+        parameters: List[str] = None,
+        dynamic_variables: List[str] = None,
+        args: List[str] = None,
         reversible: bool = None,
         check_consistency: bool = True,
-    ):
+    ) -> None:
         """Update an existing reaction.
 
-        Parameters
-        ----------
-        rate_name : str
-        function : Callable, optional
-        stoichiometry : dict, optional
-        modifiers: iterable(str), optional
-        parameters: iterable(str), optional
-        reversible: bool, optional
-
         See Also
         --------
         add_reaction
@@ -304,8 +318,8 @@ class Model(
             substrates = [k for k, v in stoichiometry.items() if v < 0]
             products = [k for k, v in stoichiometry.items() if v > 0]
         else:
-            substrates = None
-            products = None
+            substrates = None  # type: ignore
+            products = None  # type: ignore
         self.update_rate(
             rate_name=rate_name,
             function=function,
@@ -321,22 +335,22 @@ class Model(
         if stoichiometry is not None:
             self.update_stoichiometry(rate_name=rate_name, stoichiometry=stoichiometry)
 
-    def update_reactions(self, reactions: dict):
+    def update_reactions(self, reactions: dict) -> None:
         for rate_name, reaction in reactions.items():
             self.update_reaction(rate_name, **reaction)
 
     def add_reaction_from_ratelaw(
         self,
         rate_name: str,
-        ratelaw: "modelbase.ode.utils.ratelaw.BaseRateLaw",
-        **meta_info,
-    ):
+        ratelaw: Any,
+        **meta_info: Dict[str, Any],
+    ) -> None:
         """Add a reaction from a ratelaw.
 
         Parameters
         ----------
         rate_name
-        ratelaw : modelbase.ode.utils.ratelaw.BaseRateLaw
+        ratelaw
             Ratelaw instance
         meta_info
 
@@ -367,7 +381,7 @@ class Model(
         )
         self.add_stoichiometry(rate_name=rate_name, stoichiometry=ratelaw.stoichiometry)
 
-    def remove_reaction(self, rate_name: str):
+    def remove_reaction(self, rate_name: str) -> None:
         """Remove a reaction from the model.
 
         Parameters
@@ -377,7 +391,7 @@ class Model(
         self.remove_rate(rate_name=rate_name)
         self.remove_rate_stoichiometry(rate_name=rate_name)
 
-    def remove_reactions(self, rate_names: list[str]):
+    def remove_reactions(self, rate_names: List[str]) -> None:
         """Remove multiple reactions from the model.
 
         Parameters
@@ -393,9 +407,9 @@ class Model(
 
     def get_full_concentration_dict(
         self,
-        y: Union[dict[str, float], dict[str, list[float]], np.ndarray[float], list[float]],
-        t: float = 0.0,
-    ) -> dict[str, np.ndarray]:
+        y: Union[Dict[str, float], List[float], np.ndarray],
+        t: Union[float, List[float], np.ndarray] = 0.0,
+    ) -> Dict[str, np.ndarray]:
         """Calculate the derived variables (at time(s) t).
 
         Examples
@@ -411,39 +425,43 @@ class Model(
             y = {k: np.ones(len(t)) * v for k, v in y.items()}
         else:
             y = dict(zip(self.get_compounds(), (np.ones((len(t), 1)) * y).T))
-        return {k: np.ones(len(t)) * v for k, v in self._get_fcd(t=t, y=y).items()}
+        return {k: np.ones(len(t)) * v for k, v in self._get_fcd(t=t, y=y).items()}  # type: ignore
 
     def get_fluxes_dict(
         self,
-        y: Union[dict[str, float], dict[str, list[float]], np.ndarray[float], list[float]],
-        t: float = 0.0,
-    ) -> dict[str, np.ndarray]:
+        y: Union[Dict[str, float], Dict[str, List[float]], Dict[str, np.ndarray], np.ndarray, List[float]],
+        t: Union[float, List[float], np.ndarray] = 0.0,
+    ) -> Dict[str, np.ndarray]:
         """Calculate the fluxes at time point(s) t."""
         self._update_derived_parameters()
-        fcd = self.get_full_concentration_dict(y=y, t=t)
-        return {k: np.ones(len(fcd["time"])) * v for k, v in self._get_fluxes(fcd=fcd).items()}
+        fcd = self.get_full_concentration_dict(y=y, t=t)  # type: ignore
+        ones = np.ones(len(fcd["time"]))
+        if len(fcd["time"]) == 1:
+            return {k: ones * v for k, v in self._get_fluxes(fcd=fcd).items()}  # type: ignore
+        else:
+            return {k: ones * v for k, v in self._get_fluxes_array(fcd=fcd).items()}  # type: ignore
 
     def get_fluxes_array(
         self,
-        y: Union[dict[str, float], dict[str, list[float]], np.ndarray[float], list[float]],
-        t: float = 0.0,
+        y: Union[Dict[str, float], Dict[str, List[float]], Dict[str, np.ndarray], np.ndarray, List[float]],
+        t: Union[float, List[float], np.ndarray] = 0.0,
     ) -> np.ndarray:
         """Calculate the fluxes at time point(s) t."""
         return np.array([i for i in self.get_fluxes_dict(y=y, t=t).values()]).T
 
     def get_fluxes_df(
         self,
-        y: Union[dict[str, float], dict[str, list[float]], np.ndarray[float], list[float]],
-        t: float = 0.0,
+        y: Union[Dict[str, float], Dict[str, List[float]], Dict[str, np.ndarray], np.ndarray, List[float]],
+        t: Union[float, List[float], np.ndarray] = 0.0,
     ) -> pd.DataFrame:
         """Calculate the fluxes at time point(s) t."""
         if isinstance(t, (int, float)):
-            t = [t]
+            t = [t]  # type: ignore
         return pd.DataFrame(data=self.get_fluxes_dict(y=y, t=t), index=t, columns=self.get_rate_names())
 
     # This can't get keyword-only arguments, as the integrators are calling it with
     # positional arguments
-    def _get_rhs(self, t: float, y: list[float]) -> np.ndarray:
+    def _get_rhs(self, t: Union[float, List[float], np.ndarray], y: List[np.ndarray]) -> np.ndarray:
         """Calculate the right hand side of the ODE system.
 
         This is the more performant version of get_right_hand_side()
@@ -451,26 +469,26 @@ class Model(
 
         Watch out that this function swaps t and y!
         """
-        y = dict(zip(self.get_compounds(), np.array(y).reshape(-1, 1)))
-        fcd = self._get_fcd(t=t, y=y)
-        fluxes = self._get_fluxes(fcd=fcd)
+        y = dict(zip(self.get_compounds(), np.array(y).reshape(-1, 1)))  # type: ignore
+        fcd = self._get_fcd(t=t, y=y)  # type: ignore
+        fluxes = self._get_fluxes(fcd=fcd)  # type: ignore
         compounds_local = self.get_compounds()
-        dxdt = dict(zip(compounds_local, np.zeros(len(compounds_local))))
+        dxdt = dict(zip(compounds_local, it.repeat(0.0)))
         for k, stoc in self.stoichiometries_by_compounds.items():
             for flux, n in stoc.items():
                 dxdt[k] += n * fluxes[flux]
-        return np.array([dxdt[i] for i in compounds_local]).flatten()
+        return np.array([dxdt[i] for i in compounds_local], dtype="float")
 
     def get_right_hand_side(
         self,
-        y: Union[dict[str, float], dict[str, list[float]], np.ndarray[float], list[float]],
-        t: float = 0.0,
-    ) -> dict[str, float]:
+        y: Union[Dict[str, float], Dict[str, List[float]], Dict[str, np.ndarray], np.ndarray, List[float]],
+        t: Union[float, List[float], np.ndarray] = 0.0,
+    ) -> Dict[str, float]:
         """Calculate the right hand side of the ODE system."""
         self._update_derived_parameters()
-        y = self.get_full_concentration_dict(y=y, t=t)
-        y = [y[i] for i in self.get_compounds()]
-        rhs = self._get_rhs(t=t, y=y)
+        fcd = self.get_full_concentration_dict(y=y, t=t)  # type: ignore
+        fcd_array = [fcd[i] for i in self.get_compounds()]
+        rhs = self._get_rhs(t=t, y=fcd_array)
         eqs = [f"d{cpd}dt" for cpd in self.get_compounds()]
         return dict(zip(eqs, rhs))
 
@@ -478,9 +496,7 @@ class Model(
     # Model conversion functions
     ##########################################################################
 
-    def to_labelmodel(
-        self, labelcompounds: dict[str, int], labelmaps: dict[str, list[int]]
-    ) -> "modelbase.ode.LabelModel":
+    def to_labelmodel(self, labelcompounds: Dict[str, int], labelmaps: Dict[str, List[int]]) -> Any:
         """Create a LabelModel from this model.
 
         Examples
@@ -518,31 +534,29 @@ class Model(
                 parameters=module["parameters"],
             )
 
-        for rate_name, rate_dict in self.rates.items():
+        for rate_name, rate in self.rates.items():
             if rate_name not in labelmaps:
                 lm.add_reaction(
                     rate_name=rate_name,
-                    function=rate_dict["function"],
+                    function=rate["function"],
                     stoichiometry=self.stoichiometries[rate_name],
-                    modifiers=rate_dict["modifiers"],
-                    parameters=rate_dict["parameters"],
-                    reversible=rate_dict["reversible"],
+                    modifiers=rate["modifiers"],
+                    parameters=rate["parameters"],
+                    reversible=rate["reversible"],
                 )
             else:
                 lm.add_labelmap_reaction(
                     rate_name=rate_name,
-                    function=rate_dict["function"],
-                    stoichiometry=self.stoichiometries[rate_name],
+                    function=rate["function"],
+                    stoichiometry=cast(Dict[str, int], self.stoichiometries[rate_name]),
                     labelmap=labelmaps[rate_name],
-                    modifiers=rate_dict["modifiers"],
-                    parameters=rate_dict["parameters"],
-                    reversible=rate_dict["reversible"],
+                    modifiers=rate["modifiers"],
+                    parameters=rate["parameters"],
+                    reversible=rate["reversible"],
                 )
         return lm
 
-    def to_linear_labelmodel(
-        self, labelcompounds: dict[str, int], labelmaps: dict[str, list[int]]
-    ) -> "modelbase.ode.LinearLabelModel":
+    def to_linear_labelmodel(self, labelcompounds: Dict[str, int], labelmaps: Dict[str, List[int]]) -> Any:
         """Create a LinearLabelModel from this model.
 
         Watch out that for a linear label model reversible reactions have to be split
@@ -575,9 +589,9 @@ class Model(
             if compound in labelcompounds:
                 lm.add_compound(compound=compound, num_labels=labelcompounds[compound])
 
-        for rate_name, rate_dict in self.rates.items():
+        for rate_name, rate in self.rates.items():
             if rate_name in labelmaps:
-                if rate_dict["reversible"]:
+                if rate["reversible"]:
                     warnings.warn(
                         f"Reaction {rate_name} is annotated as reversible. "
                         "Did you remember to split it into a forward and backward part?"
@@ -585,7 +599,9 @@ class Model(
                 lm.add_reaction(
                     rate_name=rate_name,
                     stoichiometry={
-                        k: v for k, v in self.stoichiometries[rate_name].items() if k in labelcompounds
+                        k: v
+                        for k, v in cast(Dict[str, int], self.stoichiometries[rate_name]).items()
+                        if k in labelcompounds
                     },
                     labelmap=labelmaps[rate_name],
                 )
@@ -643,7 +659,7 @@ class Model(
     # SBML functions
     ##########################################################################
 
-    def _create_sbml_reactions(self, *, sbml_model: libsbml.Model):
+    def _create_sbml_reactions(self, *, sbml_model: libsbml.Model) -> None:
         """Create the reactions for the sbml model."""
         for rate_id, stoichiometry in self.stoichiometries.items():
             rate = self.meta_info["rates"][rate_id]
