@@ -2,23 +2,44 @@
 from __future__ import annotations
 
 import re as re
-from typing import Union, Any
+from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union, cast
 
+import matplotlib as plt
 import numpy as np
 import pandas as pd
-import matplotlib as plt
 
 from ...utils.plotting import plot, plot_grid
-from .simulator import _Simulate
+from ..integrators import AbstractIntegrator
+from ..models import LabelModel
+from .abstract_simulator import _BaseRateSimulator
 
 
-class _LabelSimulate(_Simulate):
+class _LabelSimulate(_BaseRateSimulator[LabelModel]):
     """Simulator for LabelModels."""
 
-    def __init__(self, model, integrator_name: str, **kwargs):
-        super().__init__(model=model, integrator_name=integrator_name, **kwargs)
+    def __init__(
+        self,
+        model: LabelModel,
+        integrator: Type[AbstractIntegrator],
+        y0: Optional[List[float]] = None,
+        time: Optional[List[np.ndarray]] = None,
+        results: Optional[List[np.ndarray]] = None,
+        parameters: List[Dict[str, float]] = None,
+    ) -> None:
+        super().__init__(
+            model=model,
+            integrator=integrator,
+            y0=y0,
+            time=time,
+            results=results,
+            parameters=parameters,
+        )
 
-    def generate_y0(self, base_y0: dict[str, int], label_positions: dict[str, int]) -> dict[str, float]:
+    def generate_y0(
+        self,
+        base_y0: Union[List[float], Dict[str, float]],
+        label_positions: Dict[str, Union[int, List[int]]],
+    ) -> Dict[str, float]:
         """Generate y0 for all isotopomers given a base y0.
 
         Examples
@@ -29,52 +50,82 @@ class _LabelSimulate(_Simulate):
         """
         return self.model.generate_y0(base_y0=base_y0, label_positions=label_positions)
 
-    def get_total_concentration(self, compound: str) -> np.array:
+    def get_total_concentration(self, compound: str) -> Optional[np.ndarray]:
         """Get the total concentration of all isotopomers of a compound."""
-        return self.get_full_results_dict()[compound + "__total"]
+        res = self.get_full_results_dict(concatenated=True)
+        if res is None:
+            return None
+        return cast(Dict[str, np.ndarray], res)[compound + "__total"]
 
-    def get_unlabeled_concentration(self, compound: str) -> np.array:
+    def get_unlabeled_concentration(self, compound: str) -> Optional[np.ndarray]:
         """Get the concentration of an isotopomer that is unlabeled."""
         carbons = "0" * self.model.label_compounds[compound]["num_labels"]
-        return self.get_full_results_dict()[compound + f"__{carbons}"]
+        res = self.get_full_results_dict()
+        if res is None:
+            return None
+        return cast(Dict[str, np.ndarray], res)[compound + f"__{carbons}"]
 
-    def get_total_label_concentration(self, compound: str) -> np.array:
+    def get_total_label_concentration(self, compound: str) -> Optional[np.ndarray]:
         """Get the total concentration of all labeled isotopomers of a compound."""
-        return self.get_total_concentration(compound=compound) - self.get_unlabeled_concentration(
-            compound=compound
-        )
+        total = self.get_total_concentration(compound=compound)
+        unlabeled = self.get_unlabeled_concentration(compound=compound)
+        if total is None or unlabeled is None:
+            return None
+        return cast(np.ndarray, total - unlabeled)
 
-    def get_all_isotopomer_concentrations_array(self, compound: str) -> np.ndarray:
+    def get_all_isotopomer_concentrations_array(self, compound: str) -> Optional[np.ndarray]:
         """Get concentrations of all isotopomers of a compound."""
-        return self.get_all_isotopomer_concentrations_df(compound=compound).values
+        res = self.get_all_isotopomer_concentrations_df(compound=compound)
+        if res is None:
+            return None
+        return cast(np.ndarray, res.values)
 
-    def get_all_isotopomer_concentrations_dict(self, compound: str) -> dict[str, np.ndarray]:
+    def get_all_isotopomer_concentrations_dict(self, compound: str) -> Optional[Dict[str, np.ndarray]]:
         """Get concentrations of all isotopomers of a compound."""
-        df = self.get_all_isotopomer_concentrations_df(compound=compound)
-        return dict(zip(df.columns, df.values.T))
+        res = self.get_all_isotopomer_concentrations_df(compound=compound)
+        if res is None:
+            return None
+        return dict(zip(res.columns, res.values.T))
 
-    def get_all_isotopomer_concentrations_df(self, compound: str) -> pd.Series:
+    def get_all_isotopomer_concentrations_df(self, compound: str) -> Optional[pd.Series]:
         """Get concentrations of all isotopomers of a compound."""
         isotopomers = self.model.get_compound_isotopomers(compound=compound)
-        return self.get_results_df()[isotopomers]
+        df = self.get_results_df()
+        if isotopomers is None or df is None:
+            return None
+        df = cast(pd.DataFrame, df)[isotopomers]
+        return df[isotopomers]
 
-    def get_concentrations_by_reg_exp_array(self, reg_exp: str) -> np.ndarray:
+    def get_concentrations_by_reg_exp_array(self, reg_exp: str) -> Optional[np.ndarray]:
         """Get concentrations of all isotopomers matching the regular expression."""
         isotopomers = [i for i in self.model.get_compounds() if re.match(reg_exp, i)]
-        return self.get_results_df()[isotopomers].values
+        df = self.get_results_df()
+        if isotopomers is None or df is None:
+            return None
+        df = cast(pd.DataFrame, df)[isotopomers]
+        return cast(np.ndarray, df[isotopomers].values)
 
-    def get_concentrations_by_reg_exp_dict(self, reg_exp: str) -> dict[str, np.ndarray]:
+    def get_concentrations_by_reg_exp_dict(self, reg_exp: str) -> Optional[Dict[str, np.ndarray]]:
         """Get concentrations of all isotopomers of a compound."""
         isotopomers = [i for i in self.model.get_compounds() if re.match(reg_exp, i)]
-        df = self.get_results_df()[isotopomers]
+        df = self.get_results_df(concatenated=True)
+        if isotopomers is None or df is None:
+            return None
+        df = cast(pd.DataFrame, df)[isotopomers]
         return dict(zip(df.columns, df.values.T))
 
     def get_concentrations_by_reg_exp_df(self, reg_exp: str) -> pd.DataFrame:
         """Get concentrations of all isotopomers of a compound."""
         isotopomers = [i for i in self.model.get_compounds() if re.match(reg_exp, i)]
-        return self.get_results_df()[isotopomers]
+        df = self.get_results_df(concatenated=True)
+        if isotopomers is None:
+            return None
+        df = cast(pd.DataFrame, df)[isotopomers]
+        return df[isotopomers]
 
-    def get_concentration_at_positions(self, compound: str, positions: Union[int, list[int]]) -> np.ndarray:
+    def get_concentration_at_positions(
+        self, compound: str, positions: Union[int, List[int]]
+    ) -> Optional[np.ndarray]:
         """Get concentration of an isotopomer labelled at certain position(s)."""
         if isinstance(positions, int):
             positions = [positions]
@@ -83,21 +134,37 @@ class _LabelSimulate(_Simulate):
         for position in positions:
             label_positions[position] = "1"
         reg_exp = f"{compound}__{''.join(label_positions)}"
-        return np.sum(self.get_concentrations_by_reg_exp_array(reg_exp=reg_exp), axis=1)
+        res = self.get_concentrations_by_reg_exp_array(reg_exp=reg_exp)
+        if res is None:
+            return None
+        return cast(np.ndarray, np.sum(res, axis=1))
 
-    def get_concentrations_of_n_labeled_array(self, compound: str, n_labels: int) -> np.array:
+    def get_concentrations_of_n_labeled_array(self, compound: str, n_labels: int) -> Optional[np.ndarray]:
         """Get concentrations of all isotopomers that carry n labels."""
-        return self.get_concentrations_of_n_labeled_df(compound=compound, n_labels=n_labels).values
+        res = self.get_concentrations_of_n_labeled_df(compound=compound, n_labels=n_labels)
+        if res is None:
+            return None
+        return cast(np.ndarray, res.values)
 
-    def get_concentrations_of_n_labeled_dict(self, compound: str, n_labels: int) -> dict[str, np.ndarray]:
+    def get_concentrations_of_n_labeled_dict(
+        self, compound: str, n_labels: int
+    ) -> Optional[Dict[str, np.ndarray]]:
         """Get concentrations of all isotopomers that carry n labels."""
         df = self.get_concentrations_of_n_labeled_df(compound=compound, n_labels=n_labels)
+        if df is None:
+            return None
         return dict(zip(df.columns, df.values.T))
 
-    def get_concentrations_of_n_labeled_df(self, compound: str, n_labels: int) -> pd.DataFrame:
+    def get_concentrations_of_n_labeled_df(self, compound: str, n_labels: int) -> Optional[pd.DataFrame]:
         """Get concentrations of all isotopomers that carry n labels."""
-        isotopomers = self.model.get_compound_isotopomers_with_n_labels(compound=compound, n_labels=n_labels)
-        return self.get_results_df()[isotopomers]
+        isotopomers = self.model.get_compound_isotopomers_with_n_labels(
+            compound=compound,
+            n_labels=n_labels,
+        )
+        res = self.get_results_df(concatenated=True)
+        if res is None:
+            return None
+        return cast(pd.DataFrame, res)[isotopomers]
 
     def plot(
         self,
@@ -107,21 +174,24 @@ class _LabelSimulate(_Simulate):
         grid: bool = True,
         tight_layout: bool = True,
         ax: plt.Axes = None,
-        figure_kwargs: dict[str, Any] = None,
-        subplot_kwargs: dict[str, Any] = None,
-        plot_kwargs: dict[str, Any] = None,
-        grid_kwargs: dict[str, Any] = None,
-        legend_kwargs: dict[str, Any] = None,
-        tick_kwargs: dict[str, Any] = None,
-        label_kwargs: dict[str, Any] = None,
-        title_kwargs: dict[str, Any] = None,
-    ) -> tuple[plt.Figure, plt.Axes]:
+        figure_kwargs: Dict[str, Any] = None,
+        subplot_kwargs: Dict[str, Any] = None,
+        plot_kwargs: Dict[str, Any] = None,
+        grid_kwargs: Dict[str, Any] = None,
+        legend_kwargs: Dict[str, Any] = None,
+        tick_kwargs: Dict[str, Any] = None,
+        label_kwargs: Dict[str, Any] = None,
+        title_kwargs: Dict[str, Any] = None,
+    ) -> Tuple[Optional[plt.Figure], Optional[plt.Axes]]:
         """Plot all total concentrations."""
         compounds = sorted(
             [f"{i}__total" for i in self.model.label_compounds] + self.model.nonlabel_compounds
         )
         x = self.get_time()
-        y = self.get_full_results_df().loc[:, compounds]
+        y = self.get_full_results_df(concatenated=True)
+        if x is None or y is None:
+            return None, None
+        y = cast(pd.DataFrame, y).loc[:, compounds]
         return plot(
             plot_args=(x, y),
             legend=None,
@@ -141,12 +211,16 @@ class _LabelSimulate(_Simulate):
             title_kwargs=title_kwargs,
         )
 
-    def _calculate_label_distribution(self, *, compound: str, relative: bool) -> np.ndarray:
+    def _calculate_label_distribution(self, *, compound: str, relative: bool) -> Optional[np.ndarray]:
         """Calculate the label distribution of a compound."""
         total_concentration = self.get_total_concentration(compound=compound)
+        if total_concentration is None:
+            return None
         concentrations = []
         for position in range(self.model.get_compound_number_of_label_positions(compound=compound)):
             concentration = self.get_concentration_at_positions(compound=compound, positions=position)
+            if concentration is None:
+                return None
             if relative:
                 concentration = concentration / total_concentration
             concentrations.append(concentration)
@@ -162,20 +236,22 @@ class _LabelSimulate(_Simulate):
         grid: bool = True,
         tight_layout: bool = True,
         ax: plt.Axes = None,
-        figure_kwargs: dict[str, Any] = None,
-        subplot_kwargs: dict[str, Any] = None,
-        plot_kwargs: dict[str, Any] = None,
-        grid_kwargs: dict[str, Any] = None,
-        legend_kwargs: dict[str, Any] = None,
-        tick_kwargs: dict[str, Any] = None,
-        label_kwargs: dict[str, Any] = None,
-        title_kwargs: dict[str, Any] = None,
-    ) -> tuple[plt.Figure, plt.Axes]:
+        figure_kwargs: Dict[str, Any] = None,
+        subplot_kwargs: Dict[str, Any] = None,
+        plot_kwargs: Dict[str, Any] = None,
+        grid_kwargs: Dict[str, Any] = None,
+        legend_kwargs: Dict[str, Any] = None,
+        tick_kwargs: Dict[str, Any] = None,
+        label_kwargs: Dict[str, Any] = None,
+        title_kwargs: Dict[str, Any] = None,
+    ) -> Tuple[Optional[plt.Figure], Optional[plt.Axes]]:
         """Plot label distribution of a compound."""
         if ylabel is None and relative:
             ylabel = "Relative concentration"
         x = self.get_time()
         y = self._calculate_label_distribution(compound=compound, relative=relative)
+        if x is None or y is None:
+            return None, None
         return plot(
             plot_args=(x, y),
             legend=None,
@@ -197,29 +273,31 @@ class _LabelSimulate(_Simulate):
 
     def plot_label_distribution_grid(
         self,
-        compounds: list[str],
+        compounds: List[str],
         relative: bool = True,
-        ncols: int = None,
-        sharex: bool = True,
-        sharey: bool = True,
-        xlabels: list[str] = None,
-        ylabels: list[str] = None,
-        plot_titles=None,
-        figure_title: str = None,
+        ncols: Optional[int] = None,
+        sharex: Optional[bool] = True,
+        sharey: Optional[bool] = True,
+        xlabels: Optional[Union[str, List[str]]] = None,
+        ylabels: Optional[Union[str, List[str]]] = None,
+        plot_titles: Optional[Iterable[str]] = None,
+        figure_title: Optional[str] = None,
         grid: bool = True,
         tight_layout: bool = True,
-        ax: plt.Axes = None,
-        figure_kwargs: dict[str, Any] = None,
-        subplot_kwargs: dict[str, Any] = None,
-        plot_kwargs: dict[str, Any] = None,
-        grid_kwargs: dict[str, Any] = None,
-        legend_kwargs: dict[str, Any] = None,
-        tick_kwargs: dict[str, Any] = None,
-        label_kwargs: dict[str, Any] = None,
-        title_kwargs: dict[str, Any] = None,
-    ) -> tuple[plt.Figure, np.ndarray[plt.Axes]]:
+        ax: Optional[plt.Axes] = None,
+        figure_kwargs: Optional[Dict[str, Any]] = None,
+        subplot_kwargs: Optional[Dict[str, Any]] = None,
+        plot_kwargs: Optional[Dict[str, Any]] = None,
+        grid_kwargs: Optional[Dict[str, Any]] = None,
+        legend_kwargs: Optional[Dict[str, Any]] = None,
+        tick_kwargs: Optional[Dict[str, Any]] = None,
+        label_kwargs: Optional[Dict[str, Any]] = None,
+        title_kwargs: Optional[Dict[str, Any]] = None,
+    ) -> Tuple[Optional[plt.Figure], Optional[np.ndarray]]:
         """Plot label distributions of multiple compounds on a grid."""
         time = self.get_time()
+        if time is None:
+            return None, None
         plot_groups = [
             (
                 time,
@@ -235,7 +313,6 @@ class _LabelSimulate(_Simulate):
             ylabels = "Relative concentration"
         if plot_titles is None:
             plot_titles = compounds
-
         return plot_grid(
             plot_groups=plot_groups,
             legend_groups=legend_groups,
@@ -262,27 +339,29 @@ class _LabelSimulate(_Simulate):
     def plot_all_label_distributions(
         self,
         relative: bool = True,
-        ncols: int = None,
-        sharex: bool = True,
-        sharey: bool = True,
-        xlabels: list[str] = None,
-        ylabels: list[str] = None,
-        plot_titles=None,
-        figure_title: str = None,
+        ncols: Optional[int] = None,
+        sharex: Optional[bool] = True,
+        sharey: Optional[bool] = True,
+        xlabels: Optional[Union[str, List[str]]] = None,
+        ylabels: Optional[Union[str, List[str]]] = None,
+        plot_titles: Optional[Iterable[str]] = None,
+        figure_title: Optional[str] = None,
         grid: bool = True,
         tight_layout: bool = True,
-        ax: plt.Axes = None,
-        figure_kwargs: dict[str, Any] = None,
-        subplot_kwargs: dict[str, Any] = None,
-        plot_kwargs: dict[str, Any] = None,
-        grid_kwargs: dict[str, Any] = None,
-        legend_kwargs: dict[str, Any] = None,
-        tick_kwargs: dict[str, Any] = None,
-        label_kwargs: dict[str, Any] = None,
-        title_kwargs: dict[str, Any] = None,
-    ) -> tuple[plt.Figure, np.ndarray[plt.Axes]]:
+        ax: Optional[plt.Axes] = None,
+        figure_kwargs: Optional[Dict[str, Any]] = None,
+        subplot_kwargs: Optional[Dict[str, Any]] = None,
+        plot_kwargs: Optional[Dict[str, Any]] = None,
+        grid_kwargs: Optional[Dict[str, Any]] = None,
+        legend_kwargs: Optional[Dict[str, Any]] = None,
+        tick_kwargs: Optional[Dict[str, Any]] = None,
+        label_kwargs: Optional[Dict[str, Any]] = None,
+        title_kwargs: Optional[Dict[str, Any]] = None,
+    ) -> Tuple[Optional[plt.Figure], Optional[np.ndarray]]:
         """Plot label distributions of all compounds on a grid."""
         time = self.get_time()
+        if time is None:
+            return None, None
         base_compounds = self.model.label_compounds
         plot_groups = [
             (
