@@ -1,7 +1,9 @@
 from __future__ import annotations
 
+from typing import Tuple
 
-def equilibrium(S: float, P: float, keq: float) -> tuple[float, float]:
+
+def equilibrium(S: float, P: float, keq: float) -> Tuple[float, float]:
     Total = S + P
     S = Total / (1 + keq)
     P = keq * Total / (1 + keq)
