@@ -1,27 +1,31 @@
 from __future__ import annotations
 
+from typing import List
+
+from libsbml import ASTNode
+
 from .unit_conversion import get_ast_types
 
 AST_TYPES = get_ast_types()
 
 
-def handle_ast_constant_e(node, func_arguments):
+def handle_ast_constant_e(node: ASTNode, func_arguments: List[str]) -> str:
     return "math.e"
 
 
-def handle_ast_constant_false(node, func_arguments):
+def handle_ast_constant_false(node: ASTNode, func_arguments: List[str]) -> str:
     return "False"
 
 
-def handle_ast_constant_true(node, func_arguments):
+def handle_ast_constant_true(node: ASTNode, func_arguments: List[str]) -> str:
     return "True"
 
 
-def handle_ast_constant_pi(node, func_arguments):
+def handle_ast_constant_pi(node: ASTNode, func_arguments: List[str]) -> str:
     return "math.pi"
 
 
-def handle_ast_divide(node, func_arguments):
+def handle_ast_divide(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -35,7 +39,7 @@ def handle_ast_divide(node, func_arguments):
     return " / ".join(bracketed_children)
 
 
-def handle_ast_divide_int(node, func_arguments):
+def handle_ast_divide_int(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -49,7 +53,7 @@ def handle_ast_divide_int(node, func_arguments):
     return " // ".join(bracketed_children)
 
 
-def handle_ast_function(node, func_arguments):
+def handle_ast_function(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -59,64 +63,64 @@ def handle_ast_function(node, func_arguments):
     return f"{node_name}({arguments})"
 
 
-def handle_ast_function_abs(node, func_arguments):
+def handle_ast_function_abs(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.abs({child})"
 
 
-def handle_ast_function_ceiling(node, func_arguments):
+def handle_ast_function_ceiling(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"math.ceil({child})"
 
 
-def handle_ast_function_delay(node, func_arguments):
+def handle_ast_function_delay(node: ASTNode, func_arguments: List[str]) -> str:
     raise NotImplementedError
 
 
-def handle_ast_function_exp(node, func_arguments):
+def handle_ast_function_exp(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.exp({child})"
 
 
-def handle_ast_function_factorial(node, func_arguments):
+def handle_ast_function_factorial(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"math.factorial({child})"
 
 
-def handle_ast_function_floor(node, func_arguments):
+def handle_ast_function_floor(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"math.floor({child})"
 
 
-def handle_ast_function_ln(node, func_arguments):
+def handle_ast_function_ln(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.log({child})"
 
 
-def handle_ast_function_log(node, func_arguments):
+def handle_ast_function_log(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.log10({child})"
 
 
-def handle_ast_function_max(node, func_arguments):
+def handle_ast_function_max(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
     ]
-    children = ", ".join(children)
-    return f"np.max([{children}])"
+    args = ", ".join(children)
+    return f"np.max([{args}])"
 
 
-def handle_ast_function_min(node, func_arguments):
+def handle_ast_function_min(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
     ]
-    children = ", ".join(children)
-    return f"np.min({children})"
+    args = ", ".join(children)
+    return f"np.min({args})"
 
 
-def handle_ast_function_piecewise(node, func_arguments):
+def handle_ast_function_piecewise(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -131,7 +135,7 @@ def handle_ast_function_piecewise(node, func_arguments):
         return f"({children[0]} if {children[1]})"
 
 
-def handle_ast_function_power(node, func_arguments):
+def handle_ast_function_power(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -152,29 +156,29 @@ def handle_ast_function_power(node, func_arguments):
     # return f"np.power({args})"
 
 
-def handle_ast_function_rate_of(node, func_arguments):
+def handle_ast_function_rate_of(node: ASTNode, func_arguments: List[str]) -> str:
     raise NotImplementedError
 
 
-def handle_ast_function_root(node, func_arguments):
+def handle_ast_function_root(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.sqrt({child})"
 
 
-def handle_ast_function_rem(node, func_arguments):
+def handle_ast_function_rem(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
     ]
-    children = ", ".join(children)
-    return f"np.remainder({children})"
+    args = ", ".join(children)
+    return f"np.remainder({args})"
 
 
-def handle_ast_integer(node, func_arguments):
+def handle_ast_integer(node: ASTNode, func_arguments: List[str]) -> str:
     return str(int(node.getValue()))
 
 
-def handle_ast_lambda(node, func_arguments):
+def handle_ast_lambda(node: ASTNode, func_arguments: List[str]) -> str:
     num_b_vars = node.getNumBvars()
     num_children = node.getNumChildren()
     children = [
@@ -184,7 +188,7 @@ def handle_ast_lambda(node, func_arguments):
     return ", ".join(children)
 
 
-def handle_ast_logical_and(node, func_arguments):
+def handle_ast_logical_and(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -198,11 +202,11 @@ def handle_ast_logical_and(node, func_arguments):
         return f"({args})"
 
 
-def handle_ast_logical_implies(node, func_arguments):
+def handle_ast_logical_implies(node: ASTNode, func_arguments: List[str]) -> str:
     raise NotImplementedError
 
 
-def handle_ast_logical_not(node, func_arguments):
+def handle_ast_logical_not(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -216,7 +220,7 @@ def handle_ast_logical_not(node, func_arguments):
         return f"({args})"
 
 
-def handle_ast_logical_or(node, func_arguments):
+def handle_ast_logical_or(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -230,7 +234,7 @@ def handle_ast_logical_or(node, func_arguments):
         return f"({args})"
 
 
-def handle_ast_logical_xor(node, func_arguments):
+def handle_ast_logical_xor(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -240,12 +244,12 @@ def handle_ast_logical_xor(node, func_arguments):
     elif len(children) == 1:
         return "^"
     else:
-        args = [f"({i})" for i in children]
-        args = " ^ ".join(args)
+        children = [f"({i})" for i in children]
+        args = " ^ ".join(children)
         return f"({args})"
 
 
-def handle_ast_minus(node, func_arguments):
+def handle_ast_minus(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -259,27 +263,27 @@ def handle_ast_minus(node, func_arguments):
     return " - ".join(children)
 
 
-def handle_ast_name(node, func_arguments):
-    name = node.getName()
+def handle_ast_name(node: ASTNode, func_arguments: List[str]) -> str:
+    name: str = node.getName()
     func_arguments.append(name)
     return name
 
 
-def handle_ast_name_avogadro(node, func_arguments):
+def handle_ast_name_avogadro(node: ASTNode, func_arguments: List[str]) -> str:
     return "6.02214179e+23"
     # return "6.02214076e+23"
 
 
-def handle_ast_name_time(node, func_arguments):
+def handle_ast_name_time(node: ASTNode, func_arguments: List[str]) -> str:
     func_arguments.append("time")
     return "time"
 
 
-def handle_ast_originates_in_package(node, func_arguments):
+def handle_ast_originates_in_package(node: ASTNode, func_arguments: List[str]) -> str:
     raise NotImplementedError
 
 
-def handle_ast_plus(node, func_arguments):
+def handle_ast_plus(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -287,11 +291,11 @@ def handle_ast_plus(node, func_arguments):
     return " + ".join(children)
 
 
-def handle_ast_rational(node, func_arguments):
+def handle_ast_rational(node: ASTNode, func_arguments: List[str]) -> str:
     return str(node.getValue())
 
 
-def handle_ast_real(node, func_arguments):
+def handle_ast_real(node: ASTNode, func_arguments: List[str]) -> str:
     value = str(node.getValue())
     if value == "inf":
         return "np.infty"
@@ -300,7 +304,7 @@ def handle_ast_real(node, func_arguments):
     return value
 
 
-def handle_ast_relational_eq(node, func_arguments):
+def handle_ast_relational_eq(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -310,7 +314,7 @@ def handle_ast_relational_eq(node, func_arguments):
     return f"{children[0]} == {children[1]}"
 
 
-def handle_ast_relational_geq(node, func_arguments):
+def handle_ast_relational_geq(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -320,7 +324,7 @@ def handle_ast_relational_geq(node, func_arguments):
     return f"{children[0]} >= {children[1]}"
 
 
-def handle_ast_relational_gt(node, func_arguments):
+def handle_ast_relational_gt(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -330,7 +334,7 @@ def handle_ast_relational_gt(node, func_arguments):
     return f"{children[0]} > {children[1]}"
 
 
-def handle_ast_relational_leq(node, func_arguments):
+def handle_ast_relational_leq(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -340,7 +344,7 @@ def handle_ast_relational_leq(node, func_arguments):
     return f"{children[0]} <= {children[1]}"
 
 
-def handle_ast_relational_lt(node, func_arguments):
+def handle_ast_relational_lt(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -350,7 +354,7 @@ def handle_ast_relational_lt(node, func_arguments):
     return f"{children[0]} < {children[1]}"
 
 
-def handle_ast_relational_neq(node, func_arguments):
+def handle_ast_relational_neq(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -360,7 +364,7 @@ def handle_ast_relational_neq(node, func_arguments):
     return f"{children[0]} != {children[1]}"
 
 
-def handle_ast_times(node, func_arguments):
+def handle_ast_times(node: ASTNode, func_arguments: List[str]) -> str:
     children = [
         handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
         for i in range(node.getNumChildren())
@@ -379,32 +383,32 @@ def handle_ast_times(node, func_arguments):
 ###############################################################################
 
 
-def handle_ast_trigonometric_sin(node, func_arguments):
+def handle_ast_trigonometric_sin(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.sin({child})"
 
 
-def handle_ast_trigonometric_cos(node, func_arguments):
+def handle_ast_trigonometric_cos(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.cos({child})"
 
 
-def handle_ast_trigonometric_tan(node, func_arguments):
+def handle_ast_trigonometric_tan(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.tan({child})"
 
 
-def handle_ast_trigonometric_sec(node, func_arguments):
+def handle_ast_trigonometric_sec(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.cos({child})"
 
 
-def handle_ast_trigonometric_csc(node, func_arguments):
+def handle_ast_trigonometric_csc(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.sin({child})"
 
 
-def handle_ast_trigonometric_cot(node, func_arguments):
+def handle_ast_trigonometric_cot(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.tan({child})"
 
@@ -414,32 +418,32 @@ def handle_ast_trigonometric_cot(node, func_arguments):
 ###############################################################################
 
 
-def handle_ast_trigonometric_arc_sin(node, func_arguments):
+def handle_ast_trigonometric_arc_sin(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arcsin({child})"
 
 
-def handle_ast_trigonometric_arc_cos(node, func_arguments):
+def handle_ast_trigonometric_arc_cos(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arccos({child})"
 
 
-def handle_ast_trigonometric_arc_tan(node, func_arguments):
+def handle_ast_trigonometric_arc_tan(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arctan({child})"
 
 
-def handle_ast_trigonometric_arc_cot(node, func_arguments):
+def handle_ast_trigonometric_arc_cot(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arctan(1 / ({child}))"
 
 
-def handle_ast_trigonometric_arc_sec(node, func_arguments):
+def handle_ast_trigonometric_arc_sec(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arccos(1 / ({child}))"
 
 
-def handle_ast_trigonometric_arc_csc(node, func_arguments):
+def handle_ast_trigonometric_arc_csc(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arcsin(1 / ({child}))"
 
@@ -449,32 +453,32 @@ def handle_ast_trigonometric_arc_csc(node, func_arguments):
 ###############################################################################
 
 
-def handle_ast_trigonometric_sinh(node, func_arguments):
+def handle_ast_trigonometric_sinh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.sinh({child})"
 
 
-def handle_ast_trigonometric_cosh(node, func_arguments):
+def handle_ast_trigonometric_cosh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.cosh({child})"
 
 
-def handle_ast_trigonometric_tanh(node, func_arguments):
+def handle_ast_trigonometric_tanh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.tanh({child})"
 
 
-def handle_ast_trigonometric_sech(node, func_arguments):
+def handle_ast_trigonometric_sech(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.cosh({child})"
 
 
-def handle_ast_trigonometric_csch(node, func_arguments):
+def handle_ast_trigonometric_csch(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.sinh({child})"
 
 
-def handle_ast_trigonometric_coth(node, func_arguments):
+def handle_ast_trigonometric_coth(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"1 / np.tanh({child})"
 
@@ -484,37 +488,37 @@ def handle_ast_trigonometric_coth(node, func_arguments):
 ###############################################################################
 
 
-def handle_ast_trigonometric_arc_sinh(node, func_arguments):
+def handle_ast_trigonometric_arc_sinh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arcsinh({child})"
 
 
-def handle_ast_trigonometric_arc_cosh(node, func_arguments):
+def handle_ast_trigonometric_arc_cosh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arccosh({child})"
 
 
-def handle_ast_trigonometric_arc_tanh(node, func_arguments):
+def handle_ast_trigonometric_arc_tanh(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arctanh({child})"
 
 
-def handle_ast_trigonometric_arc_csch(node, func_arguments):
+def handle_ast_trigonometric_arc_csch(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arcsinh(1 / {child})"
 
 
-def handle_ast_trigonometric_arc_sech(node, func_arguments):
+def handle_ast_trigonometric_arc_sech(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arccosh(1 / {child})"
 
 
-def handle_ast_trigonometric_arc_coth(node, func_arguments):
+def handle_ast_trigonometric_arc_coth(node: ASTNode, func_arguments: List[str]) -> str:
     child = handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
     return f"np.arctanh(1 / {child})"
 
 
-def handle_ast_node(node, func_arguments):
+def handle_ast_node(node: ASTNode, func_arguments: List[str]) -> str:
     commands = {
         "AST_CONSTANT_E": handle_ast_constant_e,
         "AST_CONSTANT_FALSE": handle_ast_constant_false,
