@@ -5,6 +5,8 @@ import math
 import re
 import warnings
 from collections import defaultdict
+from pathlib import Path
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 
 import libsbml
 import numpy as np
@@ -32,7 +34,7 @@ OPERATOR_MAPPINGS = get_operator_mappings()
 INDENT = "    "
 
 
-def _simplify_function_body(function_body):
+def _simplify_function_body(function_body: str) -> str:
     try:
         function_body = str(sympy.parse_expr(function_body))
     except AttributeError:
@@ -54,59 +56,71 @@ class Parser:
         - conversion into python / modelbase objects
     """
 
-    def __init__(self, file):
-        doc = libsbml.readSBMLFromFile(str(file))
+    def _raise_or_warn(self, msg: str, error: Type[RuntimeError]) -> None:
+        if self.raise_warnings:
+            raise error(msg)
+        else:
+            warnings.warn(msg)
+
+    def __init__(self, file: Union[str, Path], raise_warnings: bool = True) -> None:
+        self.raise_warnings = raise_warnings
+
+        if not Path(file).exists():
+            raise IOError("Model file does not exist")
+        doc: libsbml.SBML_DOCUMENT = libsbml.readSBMLFromFile(str(file))
+
         # Check for unsupported packages
         for i in range(doc.num_plugins):
             if doc.getPlugin(i).getPackageName() == "comp":
-                raise NotImplementedError("No support for comp package")
+                self._raise_or_warn("No support for comp package", NotImplementedError)
 
         self.sbml_model = doc.getModel()
         if bool(self.sbml_model.getConversionFactor()):
-            raise NotImplementedError("Conversion factors are currently not supported")
+            self._raise_or_warn("Conversion factors are currently not supported", NotImplementedError)
+
         # Collections
-        self.produced_compounds = set()
-        self.consumed_compounds = set()
-        self.derived_compounds = set()
-        self.boundary_species = set()
+        self.produced_compounds: set[str] = set()
+        self.consumed_compounds: set[str] = set()
+        self.derived_compounds: set[str] = set()
+        self.boundary_species: set[str] = set()
         # Parsed stuff
-        self.parsed_atomic_units = {}
-        self.parsed_units = {}
-        self.parsed_parameters = {}
-        self.parsed_initial_assignments = {}
-        self.parsed_compartments = {}
-        self.parsed_species = {}
-        self.parsed_functions = {}
-        self.parsed_algebraic_rules = {}
-        self.parsed_assignment_rules = {}
-        self.parsed_rate_rules = {}
-        self.parsed_reactions = {}
+        self.parsed_atomic_units: Dict[str, AtomicUnit] = {}
+        self.parsed_units: Dict[str, CompositeUnit] = {}
+        self.parsed_parameters: Dict[str, Parameter] = {}
+        self.parsed_initial_assignments: Dict[str, InitialAssignment] = {}
+        self.parsed_compartments: Dict[str, Compartment] = {}
+        self.parsed_species: Dict[str, Compound] = {}
+        self.parsed_functions: Dict[str, Function] = {}
+        self.parsed_algebraic_rules: Dict[str, AlgebraicRule] = {}
+        self.parsed_assignment_rules: Dict[str, AssignmentRule] = {}
+        self.parsed_rate_rules: Dict[str, RateRule] = {}
+        self.parsed_reactions: Dict[str, Reaction] = {}
 
         # Converted intermediates
-        self.converted_initial_assignments = {}
-        self.converted_functions = {}
-        self.converted_variables = {}
-        self.converted_constants = {}
-        self.converted_derived_constants = {}
-        self.converted_reactions = {}
-        self.converted_algebraic_modules = {}
+        self.converted_initial_assignments: Dict[str, InitialAssignment] = {}
+        self.converted_functions: Dict[str, Function] = {}
+        self.converted_variables: Dict[str, Compound] = {}
+        self.converted_constants: Dict[str, Parameter] = {}
+        self.converted_derived_constants: Dict[str, InitialAssignment] = {}
+        self.converted_reactions: Dict[str, Reaction] = {}
+        self.converted_algebraic_modules: Dict[str, AssignmentRule] = {}
 
     ###############################################################################
     # PARSING STAGE
     ###############################################################################
 
-    def parse_constraints(self):
+    def parse_constraints(self) -> None:
         if len(self.sbml_model.getListOfConstraints()) > 0:
             warnings.warn(
                 "modelbase does not support model constraints. "
                 "Check the file for hints of the range of stable solutions."
             )
 
-    def parse_events(self):
+    def parse_events(self) -> None:
         if len(self.sbml_model.getListOfEvents()) > 0:
-            raise NotImplementedError("modelbase does not currently support events.")
+            self._raise_or_warn("modelbase does not currently support events.", NotImplementedError)
 
-    def parse_units(self):
+    def parse_units(self) -> None:
         for unit_definition in self.sbml_model.getListOfUnitDefinitions():
             composite_id = unit_definition.getId()
             local_units = []
@@ -124,14 +138,14 @@ class Parser:
                 units=local_units,
             )
 
-    def parse_initial_assignments(self):
+    def parse_initial_assignments(self) -> None:
         for assignment in self.sbml_model.getListOfInitialAssignments():
             sbml_id = assignment.getId()
             node = assignment.getMath()
             if node is None:
                 continue
             sbml_math = libsbml.formulaToL3String(node)
-            parsed_arguments = []
+            parsed_arguments: List[str] = []
             function_body = handle_ast_node(node, parsed_arguments)
             parsed_arguments = sorted(set(parsed_arguments))
             self.parsed_initial_assignments[sbml_id] = InitialAssignment(
@@ -142,7 +156,7 @@ class Parser:
                 sbml_math=sbml_math,
             )
 
-    def parse_parameters(self):
+    def parse_parameters(self) -> None:
         for parameter in self.sbml_model.getListOfParameters():
             parameter_id = parameter.getId()
             self.parsed_parameters[parameter_id] = Parameter(
@@ -152,7 +166,7 @@ class Parser:
                 is_constant=parameter.getConstant(),
             )
 
-    def parse_compartments(self):
+    def parse_compartments(self) -> None:
         for compartment in self.sbml_model.getListOfCompartments():
             sbml_id = compartment.getId()
             size = compartment.getSize()
@@ -167,10 +181,10 @@ class Parser:
                 is_constant=compartment.getConstant(),
             )
 
-    def parse_species(self):
+    def parse_species(self) -> None:
         for compound in self.sbml_model.getListOfSpecies():
             if bool(compound.getConversionFactor()):
-                raise NotImplementedError("Conversion factors are not supported")
+                self._raise_or_warn("Conversion factors are not supported", NotImplementedError)
             compound_id = compound.getId()
             initial_amount = compound.getInitialAmount()
             if str(initial_amount) == "nan":
@@ -198,7 +212,7 @@ class Parser:
                 is_concentration=is_concentration,
             )
 
-    def parse_functions(self):
+    def parse_functions(self) -> None:
         for func in self.sbml_model.getListOfFunctionDefinitions():
             func_name = func.getName()
             sbml_id = func.getId()
@@ -207,7 +221,7 @@ class Parser:
             elif func_name is None or func_name == "":
                 func_name = sbml_id
 
-            parsed_arguments = []
+            parsed_arguments: List[str] = []
             node = func.getMath()
             if node is None:
                 continue
@@ -222,14 +236,14 @@ class Parser:
                 sbml_math=libsbml.formulaToL3String(node),
             )
 
-    def _parse_algebraic_rule(self, rule):
+    def _parse_algebraic_rule(self, rule: libsbml.AlgebraicRule) -> None:
         node = rule.getMath()
         if node is None:
             return None
         sbml_id = rule.getMetaId()
         sbml_math = libsbml.formulaToL3String(node)
 
-        parsed_arguments = []
+        parsed_arguments: List[str] = []
         parsed_node = handle_ast_node(node, parsed_arguments)
         parsed_arguments = sorted(set(parsed_arguments))
 
@@ -242,19 +256,19 @@ class Parser:
             function_args=[],
         )
 
-    def _parse_assignment_rule(self, rule):
+    def _parse_assignment_rule(self, rule: libsbml.AssignmentRule) -> None:
         node = rule.getMath()
         if node is None:
             return None
 
-        sbml_id = rule.getMetaId()
-        sbml_math = libsbml.formulaToL3String(node)
-        derived_compound = rule.getId()
+        sbml_id: Optional[str] = rule.getMetaId()
+        sbml_math: str = libsbml.formulaToL3String(node)
+        derived_compound: str = rule.getId()
 
         if sbml_id is None or sbml_id == "":
             sbml_id = derived_compound
 
-        parsed_arguments = []
+        parsed_arguments: List[str] = []
         parsed_node = handle_ast_node(node=node, func_arguments=parsed_arguments)
         parsed_arguments = sorted(set(parsed_arguments))
 
@@ -270,7 +284,7 @@ class Parser:
             function_args=[],
         )
 
-    def _parse_rate_rule(self, rule):
+    def _parse_rate_rule(self, rule: libsbml.RateRule) -> None:
         sbml_id = rule.getMetaId()
         derived_compound = rule.getId()
         if sbml_id is None or sbml_id == "":
@@ -280,12 +294,12 @@ class Parser:
             return None
         sbml_math = libsbml.formulaToL3String(node)
 
-        parsed_arguments = []
+        parsed_arguments: List[str] = []
         parsed_node = handle_ast_node(node=node, func_arguments=parsed_arguments)
         parsed_arguments = sorted(set(parsed_arguments))
 
         if derived_compound not in self.parsed_species:
-            raise NotImplementedError("RateRule may only target species")
+            self._raise_or_warn("RateRule may only target species", NotImplementedError)
 
         self.parsed_rate_rules[sbml_id] = RateRule(
             sbml_id=sbml_id,
@@ -297,7 +311,7 @@ class Parser:
             function_args=[],
         )
 
-    def parse_rules(self):
+    def parse_rules(self) -> None:
         """Parse rules and separate them by type"""
         for rule in self.sbml_model.getListOfRules():
             if rule.element_name == "algebraicRule":
@@ -309,7 +323,7 @@ class Parser:
             else:
                 raise ValueError("Unknown rate type")
 
-    def _parse_local_parameters(self, reaction_id, kinetic_law):
+    def _parse_local_parameters(self, reaction_id: str, kinetic_law: libsbml.KineticLaw) -> Dict[str, str]:
         """Parse local parameters"""
         parameters_to_update = {}
         for parameter in kinetic_law.getListOfLocalParameters():
@@ -325,9 +339,23 @@ class Parser:
                 value=parameter.getValue(),
                 is_constant=parameter.getConstant(),
             )
+        # Some models apparently also write local parameters in this
+        for parameter in kinetic_law.getListOfParameters():
+            old_id = parameter.getId()
+            if old_id in self.parsed_parameters:
+                new_id = f"{reaction_id}__{old_id}"
+                parameters_to_update[old_id] = new_id
+            else:
+                new_id = old_id
+            self.parsed_parameters[new_id] = Parameter(
+                sbml_id=new_id,
+                name=parameter.getName(),
+                value=parameter.getValue(),
+                is_constant=parameter.getConstant(),
+            )
         return parameters_to_update
 
-    def parse_reactions(self):
+    def parse_reactions(self) -> None:
         for reaction in self.sbml_model.getListOfReactions():
             sbml_id = reaction.getId()
             kinetic_law = reaction.getKineticLaw()
@@ -340,7 +368,7 @@ class Parser:
             is_reversible = reaction.getReversible()
             modifiers = [i.getSpecies() for i in reaction.getListOfModifiers()]
 
-            parsed_arguments = []
+            parsed_arguments: List[str] = []
             parsed_node = handle_ast_node(node=node, func_arguments=parsed_arguments)
             parsed_arguments = sorted(set(parsed_arguments))
 
@@ -352,21 +380,21 @@ class Parser:
             for i, arg in enumerate(parsed_arguments):
                 parsed_arguments[i] = parameters_to_update.get(arg, arg)
 
-            parsed_reactants = defaultdict(int)
+            parsed_reactants: defaultdict[str, int] = defaultdict(int)
             for substrate in reaction.getListOfReactants():
                 species = substrate.getSpecies()
                 if species not in self.boundary_species:
                     stoichiometry = substrate.getStoichiometry()
                     if str(stoichiometry) == "nan":
-                        raise NotImplementedError("Cannot parse stoichiometry")
+                        self._raise_or_warn("Cannot parse stoichiometry", NotImplementedError)
                     parsed_reactants[species] -= stoichiometry
-            parsed_products = defaultdict(int)
+            parsed_products: defaultdict[str, int] = defaultdict(int)
             for product in reaction.getListOfProducts():
                 species = product.getSpecies()
                 if species not in self.boundary_species:
                     stoichiometry = product.getStoichiometry()
                     if str(stoichiometry) == "nan":
-                        raise NotImplementedError("Cannot parse stoichiometry")
+                        self._raise_or_warn("Cannot parse stoichiometry", NotImplementedError)
                     parsed_products[species] += stoichiometry
 
             self.parsed_reactions[sbml_id] = Reaction(
@@ -383,7 +411,7 @@ class Parser:
                 stoichiometry={},
             )
 
-    def _parse(self):
+    def _parse(self) -> None:
         self.parse_functions()
         self.parse_units()
         self.parse_compartments()
@@ -399,7 +427,7 @@ class Parser:
     # Collection & conversion stage
     ###############################################################################
 
-    def collect_slow_changing_compounds(self):
+    def collect_slow_changing_compounds(self) -> None:
         """Needed for algebraic rules to find out which
         variable is the fast-changing on"""
         for reaction in self.parsed_reactions.values():
@@ -410,20 +438,20 @@ class Parser:
         for rate in self.parsed_rate_rules.values():
             self.produced_compounds.add(rate.derived_compound)
 
-    def _move_parameter_to_variables(self, parameter):
+    def _move_parameter_to_variables(self, parameter: Parameter) -> None:
         self.converted_variables[parameter.sbml_id] = Compound(
             sbml_id=parameter.sbml_id,
             name=parameter.name,
-            compartment="",
+            compartment=None,
             initial_amount=parameter.value,
-            substance_units=[],
+            substance_units=None,
             has_only_substance_units=True,
             has_boundary_condition=False,
             is_constant=parameter.is_constant,
             is_concentration=False,
         )
 
-    def _move_compartment_to_constants(self, compartment):
+    def _move_compartment_to_constants(self, compartment: Compartment) -> None:
         self.converted_constants[compartment.sbml_id] = Parameter(
             sbml_id=compartment.sbml_id,
             name=compartment.name,
@@ -431,7 +459,7 @@ class Parser:
             is_constant=compartment.is_constant,
         )
 
-    def _move_compartment_to_variables(self, compartment):
+    def _move_compartment_to_variables(self, compartment: Compartment) -> None:
         self.converted_variables[compartment.sbml_id] = Compound(
             sbml_id=compartment.sbml_id,
             name=compartment.name,
@@ -444,7 +472,7 @@ class Parser:
             is_concentration=False,
         )
 
-    def _move_compound_to_constants(self, compound):
+    def _move_compound_to_constants(self, compound: Compound) -> None:
         self.converted_constants[compound.sbml_id] = Parameter(
             sbml_id=compound.sbml_id,
             name=compound.name,
@@ -452,28 +480,28 @@ class Parser:
             is_constant=compound.is_constant,
         )
 
-    def convert_parameters(self):
+    def convert_parameters(self) -> None:
         for sbml_id, parameter in self.parsed_parameters.items():
             if not parameter.is_constant:
                 self._move_parameter_to_variables(parameter=parameter)
             else:
                 self.converted_constants[sbml_id] = parameter
 
-    def convert_compartments(self):
+    def convert_compartments(self) -> None:
         for compartment in self.parsed_compartments.values():
             if compartment.is_constant:
                 self._move_compartment_to_constants(compartment=compartment)
             else:
                 self._move_compartment_to_variables(compartment=compartment)
 
-    def convert_compounds(self):
+    def convert_compounds(self) -> None:
         for sbml_id, compound in self.parsed_species.items():
             if compound.is_constant:
                 self._move_compound_to_constants(compound=compound)
             else:
                 self.converted_variables[sbml_id] = compound
 
-    def _replace_function_with_operators(self, algebraic_rule):
+    def _replace_function_with_operators(self, algebraic_rule: AlgebraicRule) -> None:
         rule_body = algebraic_rule.function_body
         if "," not in rule_body:
             return None
@@ -491,7 +519,7 @@ class Parser:
 
         algebraic_rule.function_body = rule_body
 
-    def _convert_algebraic_to_assignment_rule(self, rule):
+    def _convert_algebraic_to_assignment_rule(self, rule: AlgebraicRule) -> None:
         parsed_args = rule.parsed_args
         # isolate derived_compound
         explained_args = self.produced_compounds | set(self.converted_constants)
@@ -499,9 +527,9 @@ class Parser:
         if len(unexplained_args) > 1:
             unexplained_args = list(set(unexplained_args).difference(self.consumed_compounds))
             if len(unexplained_args) > 1:
-                raise NotImplementedError("No handling for underdetermined eq")
+                self._raise_or_warn("No handling for underdetermined eq", NotImplementedError)
         elif len(unexplained_args) == 0:
-            raise NotImplementedError("No handling for already determined eq")
+            self._raise_or_warn("No handling for already determined eq", NotImplementedError)
         derived_compound = unexplained_args[0]
         parsed_args.remove(derived_compound)
 
@@ -511,9 +539,9 @@ class Parser:
             eq = sympy.parse_expr(str(rule.function_body))
             function_body = str(sympy.solve(eq, derived_compound)[0])
         except AttributeError:
-            raise NotImplementedError("Equation may not only contain symbols and operators")
+            self._raise_or_warn("Equation may not only contain symbols and operators", NotImplementedError)
         except TypeError:
-            raise NotImplementedError("Equation may not only contain symbols and operators")
+            self._raise_or_warn("Equation may not only contain symbols and operators", NotImplementedError)
 
         sbml_id = rule.sbml_id
         if sbml_id is None or sbml_id == "":
@@ -532,11 +560,11 @@ class Parser:
             function_args=[],
         )
 
-    def convert_algebraic_rules(self):
-        for rule_id, rule in self.parsed_algebraic_rules.items():
+    def convert_algebraic_rules(self) -> None:
+        for rule in self.parsed_algebraic_rules.values():
             self._convert_algebraic_to_assignment_rule(rule=rule)
 
-    def convert_assignment_rules(self):
+    def convert_assignment_rules(self) -> None:
         for sbml_id, rule in self.parsed_assignment_rules.items():
             parsed_args = rule.parsed_args
             compounds = [i for i in parsed_args if i in self.converted_variables]
@@ -557,10 +585,10 @@ class Parser:
                 function_args=function_args,
             )
 
-    def _convert_rate_rule_to_reaction(self, rule):
+    def _convert_rate_rule_to_reaction(self, rule: RateRule) -> None:
         derived_compound = rule.derived_compound
         if derived_compound in self.parsed_compartments:
-            raise NotImplementedError("Dynamic compartments are unsupported")
+            self._raise_or_warn("Dynamic compartments are unsupported", NotImplementedError)
         parsed_args = rule.parsed_args
         modifiers = rule.modifiers
         if derived_compound in parsed_args:
@@ -575,14 +603,16 @@ class Parser:
             if species.has_only_substance_units:
                 function_body = rule.function_body
             else:
-
                 compartment = self.parsed_species[derived_compound].compartment
-                if self.parsed_compartments[compartment].dimensions > 0:
-                    if self.parsed_species[derived_compound].is_concentration:
-                        function_body = rule.function_body
+                if compartment is not None:
+                    if self.parsed_compartments[compartment].dimensions > 0:
+                        if self.parsed_species[derived_compound].is_concentration:
+                            function_body = rule.function_body
+                        else:
+                            function_body = f"({rule.function_body}) * {compartment}"
+                            parsed_args.append(compartment)
                     else:
-                        function_body = f"({rule.function_body}) * {compartment}"
-                        parsed_args.append(compartment)
+                        function_body = rule.function_body
                 else:
                     function_body = rule.function_body
         except KeyError:
@@ -605,21 +635,21 @@ class Parser:
             stoichiometry={},
         )
 
-    def convert_rate_rules(self):
+    def convert_rate_rules(self) -> None:
         for rule_id, rule in self.parsed_rate_rules.items():
             self._convert_rate_rule_to_reaction(rule=rule)
 
-    def convert_reactions(self):
+    def convert_reactions(self) -> None:
         self.converted_reactions.update(copy.deepcopy(self.parsed_reactions))
 
-    def convert_initial_assignments(self):
+    def convert_initial_assignments(self) -> None:
         for target, assignment in self.parsed_initial_assignments.items():
             if target in self.converted_constants:
                 self.converted_derived_constants[target] = assignment
             else:
                 self.converted_initial_assignments[target] = assignment
 
-    def _collect_and_convert(self):
+    def _collect_and_convert(self) -> None:
         self.collect_slow_changing_compounds()
 
         # constants and variables
@@ -642,17 +672,17 @@ class Parser:
     # Handling stage
     ###############################################################################
 
-    def handle_functions(self):
+    def handle_functions(self) -> None:
         for sbml_id, function in self.parsed_functions.items():
             function.function_body = _simplify_function_body(function_body=function.function_body)
             self.converted_functions[sbml_id] = function
 
-    def _delete_derived_compounds_from_compounds(self):
+    def _delete_derived_compounds_from_compounds(self) -> None:
         for compound in self.derived_compounds:
             if compound in self.converted_variables:
                 del self.converted_variables[compound]
 
-    def _sort_assignment_rules(self):
+    def _sort_assignment_rules(self) -> None:
         """Make sure that dependencies on other assignment rules are
         calculated in order"""
         unsorted_modules = self.converted_algebraic_modules.copy()
@@ -678,36 +708,36 @@ class Parser:
 
         self.converted_algebraic_modules = sorted_modules
 
-    def _set_module_func_args(self, module):
+    def _set_module_func_args(self, module: AssignmentRule) -> None:
         parsed_args = module.parsed_args
         compounds = [i for i in parsed_args if i in self.converted_variables]
         parameters = [i for i in parsed_args if i in self.converted_constants]
         modifiers = set(module.modifiers).difference(parameters)
 
         unexplained_args = set(parsed_args).difference(compounds).difference(parameters)
-        modifiers = sorted(set(modifiers).union(unexplained_args))
+        modifiers_ = cast(List[str], sorted(set(modifiers).union(unexplained_args)))
 
-        module.function_args = compounds + modifiers + parameters
+        module.function_args = compounds + modifiers_ + parameters
         module.compounds = compounds
         module.parameters = parameters
-        module.modifiers = modifiers
+        module.modifiers = modifiers_
 
-    def handle_algebraic_modules(self):
+    def handle_algebraic_modules(self) -> None:
         self._delete_derived_compounds_from_compounds()
         for module in self.converted_algebraic_modules.values():
             # watch out for species reference
             derived_compound = module.derived_compound
             if "ref" in derived_compound:
                 if derived_compound.rsplit("ref", maxsplit=1)[0] in self.parsed_species:
-                    raise NotImplementedError("Dynamic stoichiometries not supported")
+                    self._raise_or_warn("Dynamic stoichiometries not supported", NotImplementedError)
 
             self._set_module_func_args(module=module)
             module.function_body = _simplify_function_body(function_body=module.function_body)
         self._sort_assignment_rules()
 
-    def _set_reaction_stoichiometries(self, reaction):
+    def _set_reaction_stoichiometries(self, reaction: Reaction) -> None:
         """Exclude compounds that are actually parameters"""
-        stoichiometry = {}
+        stoichiometry: Dict[str, float] = {}
         modifiers = reaction.modifiers
         for substrate, value in reaction.parsed_reactants.items():
             if substrate in self.converted_variables:
@@ -724,7 +754,7 @@ class Parser:
         reaction.stoichiometry = stoichiometry
         reaction.modifiers = sorted(set(modifiers))
 
-    def _convert_substance_amount_to_concentration(self, reaction):
+    def _convert_substance_amount_to_concentration(self, reaction: Reaction) -> None:
         """Convert mole to mole / l
         The compounds in the test are supplied in mole if has_only_substance_units is false
         In that case, the reaction equation has to be reformed like this
@@ -743,23 +773,24 @@ class Parser:
             if not species.has_only_substance_units:
                 sbml_id = species.sbml_id
                 compartment = species.compartment
-                if self.parsed_compartments[compartment].dimensions == 0:
-                    continue
-                if species.is_concentration:
-                    if compartment not in removed_compartments:
-                        pattern = f"({compartment})" + r"\b"
-                        repl = f"({compartment} / {compartment})"
+                if compartment is not None:
+                    if self.parsed_compartments[compartment].dimensions == 0:
+                        continue
+                    if species.is_concentration:
+                        if compartment not in removed_compartments:
+                            pattern = f"({compartment})" + r"\b"
+                            repl = f"({compartment} / {compartment})"
+                            function_body = re.sub(pattern, repl, function_body)
+                            removed_compartments.add(compartment)
+                    else:
+                        pattern = f"({sbml_id})" + r"\b"
+                        repl = f"({sbml_id} / {compartment})"
                         function_body = re.sub(pattern, repl, function_body)
-                        removed_compartments.add(compartment)
-                else:
-                    pattern = f"({sbml_id})" + r"\b"
-                    repl = f"({sbml_id} / {compartment})"
-                    function_body = re.sub(pattern, repl, function_body)
-                if compartment not in reaction.parsed_args:
-                    reaction.parsed_args.append(compartment)
+                    if compartment not in reaction.parsed_args:
+                        reaction.parsed_args.append(compartment)
         reaction.function_body = function_body
 
-    def _set_reaction_func_args(self, reaction):
+    def _set_reaction_func_args(self, reaction: Reaction) -> None:
         all_parameters = set(self.converted_constants)
         parsed_args = reaction.parsed_args
         # remove unused (simplify after convert to concentration)
@@ -776,19 +807,19 @@ class Parser:
             compound_args = substrates
 
         unexplained_args = set(parsed_args).difference(parameters).difference(compound_args)
-        modifiers = sorted(set(modifiers).union(unexplained_args))
-        reaction.function_args = compound_args + modifiers + parameters
+        modifiers_ = sorted(set(modifiers).union(unexplained_args))
+        reaction.function_args = compound_args + modifiers_ + parameters
         reaction.parameters = parameters
-        reaction.modifiers = modifiers
+        reaction.modifiers = modifiers_
 
-    def handle_reactions(self):
+    def handle_reactions(self) -> None:
         for sbml_id, reaction in self.converted_reactions.items():
             self._set_reaction_stoichiometries(reaction=reaction)
             self._convert_substance_amount_to_concentration(reaction=reaction)
             reaction.function_body = _simplify_function_body(function_body=reaction.function_body)
             self._set_reaction_func_args(reaction=reaction)
 
-    def handle_initial_assignments(self):
+    def handle_initial_assignments(self) -> None:
         for assignment in self.converted_initial_assignments.values():
             derived_parameter = assignment.derived_parameter
             if derived_parameter in self.parsed_species:
@@ -796,14 +827,15 @@ class Parser:
                 if species.is_concentration:
                     continue
                 compartment = self.parsed_species[derived_parameter].compartment
-                if self.parsed_compartments[compartment].dimensions == 0:
-                    continue
-                function_body = assignment.function_body
-                if compartment not in function_body:
-                    assignment.function_body = f"({function_body}) * {compartment}"
-                    assignment.function_args.append(compartment)
+                if compartment is not None:
+                    if self.parsed_compartments[compartment].dimensions == 0:
+                        continue
+                    function_body = assignment.function_body
+                    if compartment not in function_body:
+                        assignment.function_body = f"({function_body}) * {compartment}"
+                        assignment.function_args.append(compartment)
 
-    def _handle(self):
+    def _handle(self) -> None:
         self.handle_functions()
         self.handle_algebraic_modules()
         self.handle_reactions()
@@ -814,14 +846,16 @@ class Parser:
     ###############################################################################
 
     @staticmethod
-    def create_python_function(func):
+    def create_python_function(
+        func: Union[InitialAssignment, AssignmentRule, Reaction, Function]
+    ) -> Callable:
         """Create and exec function"""
         func_name = func.sbml_id
         func_args = ", ".join(func.function_args)
         func_body = func.function_body
         func_str = "\n".join(
             [
-                f"def {func_name}({func_args}):",
+                f"def {func_name}({func_args}) -> None:",
                 f"{INDENT}return {func_body}",
                 "",
             ]
@@ -830,11 +864,11 @@ class Parser:
             exec(func_str, globals(), None)
         except SyntaxError:
             raise SyntaxError(f"Invalid function definition: {func_str}")
-        func = globals()[func.sbml_id]
-        func.__source__ = func_str
-        return func
+        python_func = globals()[func.sbml_id]
+        python_func.__source__ = func_str
+        return python_func  # type: ignore
 
-    def build_model_from_sbml(self):
+    def build_model_from_sbml(self) -> Tuple[Model, Dict[str, Any]]:
         # Do the parsing
         self._parse()
         self._collect_and_convert()
@@ -871,7 +905,9 @@ class Parser:
                     parameters=assignment.function_args,
                 )
             except KeyError:
-                raise NotImplementedError("Cannot use non-parameter inputs for derived parameter")
+                self._raise_or_warn(
+                    "Cannot use non-parameter inputs for derived parameter", NotImplementedError
+                )
 
         # Compounds
         model.add_compounds(sorted(set(self.converted_variables)))
@@ -889,6 +925,10 @@ class Parser:
 
         # Reactions
         for reaction_id, reaction in self.converted_reactions.items():
+            meta_info = {
+                "sbml_function": reaction.sbml_math,
+                "python_function": reaction.function_body,
+            }
             model.add_reaction(
                 rate_name=reaction_id,
                 function=self.create_python_function(func=reaction),
@@ -896,8 +936,7 @@ class Parser:
                 modifiers=reaction.modifiers,
                 parameters=reaction.parameters,
                 reversible=reaction.is_reversible,
-                sbml_function=reaction.sbml_math,
-                python_function=reaction.function_body,
+                **meta_info,  # type: ignore
             )
 
         y0 = []
@@ -905,20 +944,21 @@ class Parser:
             initial_amount = i.initial_amount
             if i.has_only_substance_units:
                 if i.is_concentration:
-                    initial_amount *= self.parsed_compartments[i.compartment].size
+                    if i.compartment is not None:
+                        initial_amount *= self.parsed_compartments[i.compartment].size
             y0.append(initial_amount)
-        y0 = np.array(y0)
-        y0[np.isnan(y0)] = 0
-        y0 = dict(zip(self.converted_variables, y0))
+        y0_arr = np.array(y0)
+        y0_arr[np.isnan(y0)] = 0
+        y0_dict = dict(zip(self.converted_variables, y0))
 
         # Initial assignments
-        possible_sources = {**model.get_full_concentration_dict(y0), **model.parameters}
+        possible_sources: Dict[str, Any] = {**model.get_full_concentration_dict(y0), **model.parameters}
         for par_name, assignment in self.converted_initial_assignments.items():
-            func = self.create_python_function(func=assignment)
-            res = func(*[possible_sources[k] for k in assignment.function_args])
+            py_func = self.create_python_function(func=assignment)
+            res = py_func(*[possible_sources[k] for k in assignment.function_args])
             if isinstance(res, (list, np.ndarray)):
                 res = res[0]
             if par_name not in model.compounds and par_name not in model.get_derived_compounds():
-                raise NotImplementedError("Initial assignment targeting unknown type")
-            y0[par_name] = res
-        return model, y0
+                self._raise_or_warn("Initial assignment targeting unknown type", NotImplementedError)
+            y0_dict[par_name] = res
+        return model, y0_dict
