@@ -1,5 +1,7 @@
 from __future__ import annotations
 
+from typing import Dict
+
 import libsbml
 
 unit_conversion = {
@@ -43,14 +45,14 @@ unit_conversion = {
 }
 
 
-def get_unit_conversion():
+def get_unit_conversion() -> Dict[str, str]:
     return {getattr(libsbml, k): v.lower() for k, v in unit_conversion.items()}
 
 
 ast_types = dict(sorted({getattr(libsbml, i): i for i in dir(libsbml) if i.startswith("AST_")}.items()))
 
 
-def get_ast_types():
+def get_ast_types() -> Dict[str, str]:
     return ast_types
 
 
@@ -63,5 +65,5 @@ operator_mappings = {
 }
 
 
-def get_operator_mappings():
+def get_operator_mappings() -> Dict[str, str]:
     return operator_mappings
