@@ -3,26 +3,26 @@ from __future__ import annotations
 
 import itertools as it
 import math
-from matplotlib.collections import QuadMesh
-import pandas as pd
-from typing import Any
+from typing import Any, Dict, Iterable, List, Optional, Sized, Tuple, Union
 
 import matplotlib.pyplot as plt
 import numpy as np
+import pandas as pd
+from matplotlib.collections import QuadMesh
 from matplotlib.colors import LogNorm, Normalize, SymLogNorm
 from matplotlib.colors import colorConverter as _colorConverter
 
 
 def _get_plot_kwargs(
-    figure_kwargs: dict[str, Any],
-    subplot_kwargs: dict[str, Any],
-    plot_kwargs: dict[str, Any],
-    grid_kwargs: dict[str, Any],
-    tick_kwargs: dict[str, Any],
-    label_kwargs: dict[str, Any],
-    title_kwargs: dict[str, Any],
-    legend_kwargs: dict[str, Any],
-) -> dict[str, dict[str, Any]]:
+    figure_kwargs: Optional[Dict[str, Any]],
+    subplot_kwargs: Optional[Dict[str, Any]],
+    plot_kwargs: Optional[Dict[str, Any]],
+    grid_kwargs: Optional[Dict[str, Any]],
+    tick_kwargs: Optional[Dict[str, Any]],
+    label_kwargs: Optional[Dict[str, Any]],
+    title_kwargs: Optional[Dict[str, Any]],
+    legend_kwargs: Optional[Dict[str, Any]],
+) -> Dict[str, Dict[str, Any]]:
     """Get default plot kwargs or overwrite them."""
     local_figure_kwargs = {"figsize": (10, 7)}
     local_subplot_kwargs = {"facecolor": "white"}
@@ -91,7 +91,7 @@ def _style_subplot(
     title: str = None,
     legend: bool = False,
     grid: bool = None,
-):
+) -> None:
     """Set style of the subplot."""
     if grid:
         ax.grid(True, which="major", axis="both", **kwargs["grid"])
@@ -111,23 +111,23 @@ def _style_subplot(
 
 
 def plot(
-    plot_args: tuple[list[float], list[float]],
-    legend: list[str] = None,
-    xlabel: str = None,
-    ylabel: str = None,
-    title: str = None,
+    plot_args: Iterable,
+    legend: Optional[Iterable[str]] = None,
+    xlabel: Optional[str] = None,
+    ylabel: Optional[str] = None,
+    title: Optional[str] = None,
     grid: bool = True,
     tight_layout: bool = True,
-    ax: plt.Axes = None,
-    figure_kwargs: dict[str, Any] = None,
-    subplot_kwargs: dict[str, Any] = None,
-    plot_kwargs: dict[str, Any] = None,
-    grid_kwargs: dict[str, Any] = None,
-    legend_kwargs: dict[str, Any] = None,
-    tick_kwargs: dict[str, Any] = None,
-    label_kwargs: dict[str, Any] = None,
-    title_kwargs: dict[str, Any] = None,
-) -> tuple[plt.Figure, plt.Axes]:
+    ax: Optional[plt.Axes] = None,
+    figure_kwargs: Optional[Dict[str, Any]] = None,
+    subplot_kwargs: Optional[Dict[str, Any]] = None,
+    plot_kwargs: Optional[Dict[str, Any]] = None,
+    grid_kwargs: Optional[Dict[str, Any]] = None,
+    legend_kwargs: Optional[Dict[str, Any]] = None,
+    tick_kwargs: Optional[Dict[str, Any]] = None,
+    label_kwargs: Optional[Dict[str, Any]] = None,
+    title_kwargs: Optional[Dict[str, Any]] = None,
+) -> Tuple[plt.Figure, plt.Axes]:
     """Plot simulation results as a grid."""
     kwargs = _get_plot_kwargs(
         figure_kwargs=figure_kwargs,
@@ -169,27 +169,27 @@ def plot(
 
 
 def plot_grid(
-    plot_groups: list[tuple[list[float], list[float]]],
-    legend_groups: list[str] = None,
-    ncols: int = None,
-    sharex: bool = True,
-    sharey: bool = True,
-    xlabels: list[str] = None,
-    ylabels: list[str] = None,
+    plot_groups: Sized,
+    legend_groups: Optional[Iterable] = None,
+    ncols: Optional[int] = None,
+    sharex: Optional[bool] = True,
+    sharey: Optional[bool] = True,
+    xlabels: Optional[Union[str, Iterable[str]]] = None,
+    ylabels: Optional[Union[str, Iterable[str]]] = None,
     figure_title: str = None,
-    plot_titles=None,
+    plot_titles: Optional[Union[Iterable[str], Iterable[None]]] = None,
     grid: bool = True,
     tight_layout: bool = True,
-    ax: plt.Axes = None,
-    figure_kwargs: dict[str, Any] = None,
-    subplot_kwargs: dict[str, Any] = None,
-    plot_kwargs: dict[str, Any] = None,
-    grid_kwargs: dict[str, Any] = None,
-    legend_kwargs: dict[str, Any] = None,
-    tick_kwargs: dict[str, Any] = None,
-    label_kwargs: dict[str, Any] = None,
-    title_kwargs: dict[str, Any] = None,
-) -> tuple[plt.Figure, list[plt.Axes]]:
+    ax: Optional[plt.Axes] = None,
+    figure_kwargs: Optional[Dict[str, Any]] = None,
+    subplot_kwargs: Optional[Dict[str, Any]] = None,
+    plot_kwargs: Optional[Dict[str, Any]] = None,
+    grid_kwargs: Optional[Dict[str, Any]] = None,
+    legend_kwargs: Optional[Dict[str, Any]] = None,
+    tick_kwargs: Optional[Dict[str, Any]] = None,
+    label_kwargs: Optional[Dict[str, Any]] = None,
+    title_kwargs: Optional[Dict[str, Any]] = None,
+) -> Tuple[plt.Figure, np.ndarray]:
     """Plot simulation results as a grid."""
     if ncols is None:
         _, ncols = min(
@@ -240,7 +240,7 @@ def plot_grid(
         subplot_kw=kwargs["subplot"],
         **kwargs["figure"],
     )
-    for ax, plot_args, legend_args, title in zip(axs.ravel(), plot_groups, legend_groups, plot_titles):
+    for ax, plot_args, legend_args, title in zip(axs.ravel(), plot_groups, legend_groups, plot_titles):  # type: ignore
         ax.plot(*plot_args, **kwargs["plot"])
         ax.set_title(title, **kwargs["title"])
         if legend_args is not None:
@@ -267,7 +267,7 @@ def plot_grid(
     return fig, axs
 
 
-def relative_luminance(color: list[float]) -> float:
+def relative_luminance(color: List[float]) -> float:
     """Calculate the relative luminance of a color."""
     rgb = _colorConverter.to_rgba_array(color)[:, :3]
 
@@ -275,11 +275,11 @@ def relative_luminance(color: list[float]) -> float:
     rsrgb = np.where(rgb <= 0.03928, rgb / 12.92, ((rgb + 0.055) / 1.055) ** 2.4)
 
     # L = 0.2126 * R + 0.7152 * G + 0.0722 * B
-    rel_luminance = np.matmul(rsrgb, [0.2126, 0.7152, 0.0722])
+    rel_luminance: List[float] = np.matmul(rsrgb, [0.2126, 0.7152, 0.0722])
     return rel_luminance[0]
 
 
-def get_norm(vmin: float, vmax: float):
+def get_norm(vmin: float, vmax: float) -> plt.Normalize:
     if vmax < 1000 and vmin > -1000:
         norm = Normalize(vmin=vmin, vmax=vmax)
     elif vmin <= 0:
@@ -302,7 +302,7 @@ def heatmap_from_dataframe(
     norm: plt.Normalize = None,
     ax: plt.Axes = None,
     cax: plt.Axes = None,
-) -> tuple[plt.Figure, plt.Axes, QuadMesh]:
+) -> Tuple[plt.Figure, plt.Axes, QuadMesh]:
     data = df.values
     rows = df.index
     columns = df.columns
