@@ -1,11 +1,15 @@
+# type: ignore
+
 from __future__ import annotations
 
+from typing import List
+
 import pytest
 
 
-def _multiline_comparison(expected: list[str], test: str):
+def _multiline_comparison(expected: List[str], test: str):
     for l1, l2 in zip(expected, test.split("\n")):
-        assert l1 == l2.strip("\r")  # for windows
+        assert l1 == l2
 
 
 @pytest.fixture()
