@@ -1,4 +1,5 @@
-import pytest
+# type: ignore
+
 import unittest
 
 from modelbase.core.utils import (
@@ -12,23 +13,23 @@ from modelbase.core.utils import (
     get_function_source_code,
 )
 
+import pytest
+
 
 def test_get_function_source_code(multiline_comparison):
     def func(a):
-        return a  # pragma: no cover
+        return a
 
-    multiline_comparison(
-        ["    def func(a):", "        return a  # pragma: no cover"], get_function_source_code(func)
-    )
+    multiline_comparison(["    def func(a):", "        return a"], get_function_source_code(func))
 
 
 def test_get_function_source_code_exec(multiline_comparison):
     exec("def func(a):\n    return a")
     function = locals()["func"]
-    function.__source__ = "def func(a):\n    return a  # pragma: no cover"
+    function.__source__ = "def func(a):\n    return a"
 
     multiline_comparison(
-        ["def func(a):", "    return a  # pragma: no cover"],
+        ["def func(a):", "    return a"],
         get_function_source_code(function),
     )
 
@@ -42,10 +43,10 @@ def test_get_formatted_function_source_code_fail_on_wrong_type():
 
 def test_get_formatted_function_source_code(multiline_comparison):
     def func(a):
-        return a  # pragma: no cover
+        return a
 
     multiline_comparison(
-        ("def func(a):", "    return a  # pragma: no cover"),
+        ("def func(a):", "    return a"),
         get_formatted_function_source_code(function_name="func", function=func, function_type="rate"),
     )
 
@@ -53,10 +54,10 @@ def test_get_formatted_function_source_code(multiline_comparison):
 def test_get_formatted_function_source_code_exec(multiline_comparison):
     exec("def func(a):\n    return a")
     function = locals()["func"]
-    function.__source__ = "def func(a):\n    return a  # pragma: no cover"
+    function.__source__ = "def func(a):\n    return a"
 
     multiline_comparison(
-        ("def func(a):", "    return a  # pragma: no cover"),
+        ("def func(a):", "    return a"),
         get_formatted_function_source_code(function_name="func", function=function, function_type="rate"),
     )
 
