@@ -1,17 +1,18 @@
+# type: ignore
+
 import unittest
-import pytest
 
 import numpy as np
+import pytest
 from modelbase.ode import Model
 from modelbase.ode import ratelaws as rl
 
 
-
 def test_str(multiline_comparison):
     model = Model()
     model.add_compounds(("x", "y"))
     model.add_stoichiometry("v1", {"x": -1, "y": 1})
-    multiline_comparison(("Model:","    2 Compounds","    1 Reactions"), str(model))
+    multiline_comparison(("Model:", "    2 Compounds", "    1 Reactions"), str(model))
 
 
 class ModelBaseTests(unittest.TestCase):
@@ -1110,9 +1111,12 @@ class ModelConversionTests(unittest.TestCase):
 
         lm = model.to_linear_labelmodel(labelcompounds=labelcompounds, labelmaps=labelmaps)
         self.assertEqual(lm.compounds, ["X__0", "Y__0", "Z__0"])
-        self.assertEqual(lm.rates, {"v1__0": {"base_name": "v1", "substrate": "X__0"}})
         self.assertEqual(lm.stoichiometries, {"v1__0": {"X__0": -1, "Y__0": 1}})
 
+        rate = lm.rates["v1__0"]
+        self.assertEqual(rate["base_name"], "v1")
+        self.assertEqual(rate["substrate"], "X__0")
+
     def test_to_linear_labelmodel_no_label_info(self):
         model = Model()
         model.add_compounds(("X", "Y", "ATP", "ADP", "Z"))
@@ -1151,9 +1155,11 @@ class ModelConversionTests(unittest.TestCase):
 
         with self.assertWarns(UserWarning):
             lm = model.to_linear_labelmodel(labelcompounds=labelcompounds, labelmaps=labelmaps)
-            self.assertEqual(lm.compounds, ["X__0", "Y__0", "Z__0"])
-            self.assertEqual(lm.rates, {"v1__0": {"base_name": "v1", "substrate": "X__0"}})
             self.assertEqual(lm.stoichiometries, {"v1__0": {"X__0": -1, "Y__0": 1}})
+            self.assertEqual(lm.compounds, ["X__0", "Y__0", "Z__0"])
+            rate = lm.rates["v1__0"]
+            self.assertEqual(rate["base_name"], "v1")
+            self.assertEqual(rate["substrate"], "X__0")
 
 
 class SBMLTests(unittest.TestCase):
@@ -1227,3 +1233,26 @@ class SBMLTests(unittest.TestCase):
         )
         with self.assertWarns(UserWarning):
             model._model_to_sbml()
+
+
+# New tests
+def test_add_reaction_from_args():
+    parameters = {"k": 1}
+    model = Model(parameters=parameters)
+    model.add_compounds(["x", "y", "xi"])
+    model.add_reaction_from_args(
+        rate_name="v1",
+        function=lambda x, y, time, k: k / time * (x - y),
+        stoichiometry={"x": -1, "y": 1},
+        args=["x", "y", "time", "k"],
+        reversible=True,
+    )
+    rate = model.rates["v1"]
+    assert rate["parameters"] == ["k"]
+    assert rate["substrates"] == ["x"]
+    assert rate["products"] == ["y"]
+    assert rate["modifiers"] == ["time"]
+    assert rate["dynamic_variables"] == ["x", "y", "time"]
+    assert rate["reversible"]
+    assert model.stoichiometries == {"v1": {"x": -1, "y": 1}}
+    assert model.stoichiometries_by_compounds == {"x": {"v1": -1}, "y": {"v1": 1}}
