@@ -1,3 +1,4 @@
+# type: ignore
 
 import unittest
 from tempfile import TemporaryDirectory
@@ -94,11 +95,6 @@ class SimulatorBaseTests(unittest.TestCase):
         s = Simulator(model=model)
         self.assertEqual(s.model, model)
 
-    def test_init_wrong_name(self):
-        model = Model()
-        with self.assertRaises(NotImplementedError):
-            Simulator(model=model, integrator_name="something-other")
-
     def test_init_wrong_model_type(self):
         class TestModel:
             pass
@@ -400,6 +396,5 @@ class SimulationTests(unittest.TestCase):
 def test_simulate_warn_no_rates_set():
     model = Model()
     sim = Simulator(model)
-    sim.initialise(y0={})
-    with pytest.warns(UserWarning):
-        sim.simulate(t_end=1)
+    with pytest.raises(AttributeError):
+        sim.initialise(y0={})
