@@ -1,11 +1,13 @@
+# type: ignore
 
 import unittest
 
-import matplotlib.pyplot as plt
-import numpy as np
 from modelbase.ode import Model, mca
 from modelbase.ode import ratefunctions as rf
 
+import matplotlib.pyplot as plt
+import numpy as np
+
 
 class MCATests(unittest.TestCase):
     def testSubstrateElasticityIrreversibleMassAction(self):
@@ -127,18 +129,22 @@ class MCALinearChainTests(unittest.TestCase):
         np.testing.assert_array_almost_equal(
             mca.get_compound_elasticity(model=model, compound="X", y=y0, t=0, normalized=False),
             [0.0, 1.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_compound_elasticity(model=model, compound="X", y=y0, t=0, normalized=True),
             [0.0, 1.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_compound_elasticity(model=model, compound="Y", y=y0, t=0, normalized=False),
             [0.0, 0.0, 1.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_compound_elasticity(model=model, compound="Y", y=y0, t=0, normalized=True),
             [0.0, 0.0, 1.0],
+            decimal=4,
         )
 
     def test_get_compound_elasticities_array(self):
@@ -152,6 +158,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=False,
             ),
             [[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_compound_elasticities_array_normalized(self):
@@ -165,6 +172,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=True,
             ),
             [[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_compound_elasticities_df(self):
@@ -197,6 +205,7 @@ class MCALinearChainTests(unittest.TestCase):
         np.testing.assert_array_almost_equal(
             df.values,
             [[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_parameter_elasticity(self):
@@ -210,6 +219,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=False,
             ),
             [1.0, 0.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_parameter_elasticity(
@@ -220,6 +230,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=True,
             ),
             [1.0, 0.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_parameter_elasticity(
@@ -230,6 +241,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=False,
             ),
             [0.0, 1.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_parameter_elasticity(
@@ -240,6 +252,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=True,
             ),
             [0.0, 1.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_parameter_elasticity(
@@ -250,6 +263,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=False,
             ),
             [0.0, 0.0, 1.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_parameter_elasticity(
@@ -260,6 +274,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=True,
             ),
             [0.0, 0.0, 1.0],
+            decimal=4,
         )
 
     def test_get_parameter_elasticities_array(self):
@@ -273,6 +288,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=False,
             ),
             [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_parameter_elasticities_array_normalized(self):
@@ -286,6 +302,7 @@ class MCALinearChainTests(unittest.TestCase):
                 normalized=True,
             ),
             [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_parameter_elasticities_df(self):
@@ -302,6 +319,7 @@ class MCALinearChainTests(unittest.TestCase):
         np.testing.assert_array_almost_equal(
             df.values,
             [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_parameter_elasticities_df_normalized(self):
@@ -318,6 +336,7 @@ class MCALinearChainTests(unittest.TestCase):
         np.testing.assert_array_almost_equal(
             df.values,
             [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
+            decimal=4,
         )
 
     def test_get_concentration_response_coefficient(self):
@@ -327,60 +346,60 @@ class MCALinearChainTests(unittest.TestCase):
                 model=model,
                 parameter="k_in",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [1, 1],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_concentration_response_coefficient(
                 model=model,
                 parameter="k_in",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [1, 1],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_concentration_response_coefficient(
                 model=model,
                 parameter="k_fwd",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [-1, 0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_concentration_response_coefficient(
                 model=model,
                 parameter="k_fwd",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [-1, 0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_concentration_response_coefficient(
                 model=model,
                 parameter="k_out",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [0, -1],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_concentration_response_coefficient(
                 model=model,
                 parameter="k_out",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [0, -1],
+            decimal=4,
         )
 
     def test_get_concentration_response_coefficients_array(self):
@@ -389,7 +408,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
         np.testing.assert_array_almost_equal(
@@ -399,6 +417,7 @@ class MCALinearChainTests(unittest.TestCase):
                 [-1, 0],
                 [0, -1],
             ],
+            decimal=4,
         )
 
     def test_get_concentration_response_coefficients_array_normalized(self):
@@ -407,7 +426,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=True,
         )
         np.testing.assert_array_almost_equal(
@@ -417,6 +435,7 @@ class MCALinearChainTests(unittest.TestCase):
                 [-1, 0],
                 [0, -1],
             ],
+            decimal=4,
         )
 
     def test_get_concentration_response_coefficients_df(self):
@@ -425,7 +444,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
         np.testing.assert_array_almost_equal(
@@ -435,6 +453,7 @@ class MCALinearChainTests(unittest.TestCase):
                 [-1, 0],
                 [0, -1],
             ],
+            decimal=4,
         )
         self.assertEqual(list(df.index), ["k_in", "k_fwd", "k_out"])
         self.assertEqual(list(df.columns), ["X", "Y"])
@@ -445,7 +464,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=True,
         )
         np.testing.assert_array_almost_equal(
@@ -455,6 +473,7 @@ class MCALinearChainTests(unittest.TestCase):
                 [-1, 0],
                 [0, -1],
             ],
+            decimal=4,
         )
         self.assertEqual(list(df.index), ["k_in", "k_fwd", "k_out"])
         self.assertEqual(list(df.columns), ["X", "Y"])
@@ -466,60 +485,60 @@ class MCALinearChainTests(unittest.TestCase):
                 model=model,
                 parameter="k_in",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [1, 1, 1],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_flux_response_coefficient(
                 model=model,
                 parameter="k_in",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [1, 1, 1],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_flux_response_coefficient(
                 model=model,
                 parameter="k_fwd",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [0.0, 0.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_flux_response_coefficient(
                 model=model,
                 parameter="k_fwd",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [0.0, 0.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_flux_response_coefficient(
                 model=model,
                 parameter="k_out",
                 y=y0,
-                t=0,
                 normalized=False,
             ),
             [0.0, 0.0, 0.0],
+            decimal=4,
         )
         np.testing.assert_array_almost_equal(
             mca.get_flux_response_coefficient(
                 model=model,
                 parameter="k_out",
                 y=y0,
-                t=0,
                 normalized=True,
             ),
             [0.0, 0.0, 0.0],
+            decimal=4,
         )
 
     def test_get_flux_response_coefficients_array(self):
@@ -528,10 +547,13 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
-        np.testing.assert_array_almost_equal(arr, [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
+        np.testing.assert_array_almost_equal(
+            arr,
+            [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
+            decimal=4,
+        )
 
     def test_get_flux_response_coefficients_array_normalized(self):
         model, y0, parameters, compounds = self.create_model()
@@ -539,10 +561,13 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
-        np.testing.assert_array_almost_equal(arr, [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
+        np.testing.assert_array_almost_equal(
+            arr,
+            [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
+            decimal=4,
+        )
 
     def test_get_flux_response_coefficients_df(self):
         model, y0, parameters, compounds = self.create_model()
@@ -550,12 +575,15 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
         self.assertEqual(list(df.index), ["k_in", "k_fwd", "k_out"])
         self.assertEqual(list(df.columns), ["v_in", "v1", "v_out"])
-        np.testing.assert_array_almost_equal(df.values, [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
+        np.testing.assert_array_almost_equal(
+            df.values,
+            [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
+            decimal=4,
+        )
 
     def test_get_flux_response_coefficients_df_normalized(self):
         model, y0, parameters, compounds = self.create_model()
@@ -563,12 +591,15 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=True,
         )
         self.assertEqual(list(df.index), ["k_in", "k_fwd", "k_out"])
         self.assertEqual(list(df.columns), ["v_in", "v1", "v_out"])
-        np.testing.assert_array_almost_equal(df.values, [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
+        np.testing.assert_array_almost_equal(
+            df.values,
+            [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
+            decimal=4,
+        )
 
     def test_plot_concentration_response_coefficients(self):
         model, y0, parameters, compounds = self.create_model()
@@ -576,7 +607,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
         plt.close()
@@ -587,7 +617,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=True,
         )
         plt.close()
@@ -598,7 +627,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=False,
         )
         plt.close()
@@ -609,7 +637,6 @@ class MCALinearChainTests(unittest.TestCase):
             model=model,
             parameters=parameters,
             y=y0,
-            t=0,
             normalized=True,
         )
         plt.close()
